# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 15:05:13 2023

@author: Wanxiang Shen
"""

from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F


class TripletLoss(nn.Module):
    def __init__(self, margin=1.0, metric="cosine"):
        super(TripletLoss, self).__init__()
        self.margin = margin
        self.metric = metric
        self.name = "TripletLoss"

    def calc_euclidean(self, x1, x2):
        return (x1 - x2).pow(2).sum(1)

    def forward(
        self, anchor: torch.Tensor, positive: torch.Tensor, negative: torch.Tensor
    ) -> torch.Tensor:

        if self.metric == "euclidean":
            distance_positive = self.calc_euclidean(anchor, positive)
            distance_negative = self.calc_euclidean(anchor, negative)
        elif self.metric == "cosine":
            distance_positive = 1 - F.cosine_similarity(anchor, positive)
            distance_negative = 1 - F.cosine_similarity(anchor, negative)

        else:
            raise ValueError("Unsupported metric: {}".format(self.metric))

        losses = torch.relu(distance_positive - distance_negative + self.margin)

        # print(distance_positive.mean())

        return losses.mean()


class TriSimplexLoss(nn.Module):

    def __init__(self):
        super(TriSimplexLoss, self).__init__()
        self.name = "TriSimplexLoss"

    def forward(self, trisimplex_input, trisimplex_emb):
        a, p, n = trisimplex_input
        a_emb, p_emb, n_emb = trisimplex_emb
        # s1 = F.cosine_similarity(a, p)
        # s2 = F.cosine_similarity(a, n)
        # s3 = F.cosine_similarity(n, p)
        # s1_emb = F.cosine_similarity(a_emb, p_emb)
        # s2_emb = F.cosine_similarity(a_emb, n_emb)
        # s3_emb = F.cosine_similarity(n_emb, p_emb)
        # loss = F.mse_loss(s1, s1_emb) + F.mse_loss(s2, s2_emb) + F.mse_loss(s3, s3_emb)

        sim_a = self._pairwise_cosine_sim(a)
        sim_a_emb = self._pairwise_cosine_sim(a_emb)

        loss = F.mse_loss(sim_a, sim_a_emb)
        return loss

    def _pairwise_cosine_sim(self, tensor):
        # Normalize each vector in the tensor
        normalized_tensor = tensor / tensor.norm(dim=1, keepdim=True)
        # Perform matrix multiplication with its transpose to get cosine similarity
        similarity_matrix = torch.matmul(normalized_tensor, normalized_tensor.T)

        upper_triangular = torch.triu(similarity_matrix, diagonal=1)
        indices = upper_triangular != 0
        non_squareform_vector = upper_triangular[indices]
        return non_squareform_vector


class MAEWithNaNLabelsLoss(nn.Module):
    def __init__(self):
        super(MAEWithNaNLabelsLoss, self).__init__()

    def forward(self, predictions, labels):
        mask = ~torch.isnan(labels)[:, 0]  # Assuming labels is a 2D tensor
        if mask.any():
            # mae_loss = torch.mean((predictions[mask] - labels[mask]) ** 2)
            mae_loss = torch.mean(torch.abs(predictions[mask] - labels[mask]))

            return mae_loss
        else:
            return torch.tensor(0.0, device=labels.device, dtype=labels.dtype)



class CEWithNaNLabelsLoss(nn.Module):
    def __init__(self, weights=None):
        super().__init__()
        self.weights = weights

    def forward(self, logits, labels):
        """
        logits: (B, C) raw logits
        labels: (B, C) one-hot / soft labels, may contain NaN
        """
        # mask rows with any NaN
        mask = ~torch.isnan(labels).any(dim=1)
        if not mask.any():
            return torch.tensor(0.0, device=labels.device)

        y_pred = logits[mask]                 # (B, C)
        y_true = labels[mask].argmax(dim=1)   # (B,)

        if self.weights is None:
            return F.cross_entropy(y_pred, y_true)
        else:
            w = torch.tensor(
                self.weights,
                device=y_true.device,
                dtype=torch.float,
            )
            return F.cross_entropy(y_pred, y_true, weight=w)




class FocalLoss(nn.Module):
    """Focal loss for class imbalance (Lin et al., 2020)."""

    def __init__(self, weights=None, gamma=2.0, reduction="mean"):
        super(FocalLoss, self).__init__()
        alpha = weights
        self.gamma = gamma
        self.reduction = reduction
        self.alpha = alpha
        if isinstance(alpha, (float, int, list)):
            self.alpha = torch.tensor(alpha, dtype=torch.float)

    def forward(self, predictions, labels):
        if labels.dim() > 1:
            labels = labels.argmax(dim=1)
        pt = torch.gather(predictions, 1, labels.unsqueeze(1)).squeeze()
        if self.alpha is not None:
            if self.alpha.type() != predictions.data.type():
                self.alpha = self.alpha.type_as(predictions.data)
            at = torch.gather(self.alpha, 0, labels)
        else:
            at = 1.0
        log_pt = torch.log(pt)
        loss = -at * ((1 - pt) ** self.gamma) * log_pt
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        else:
            return loss


class DiceLoss(nn.Module):
    def __init__(self, epsilon=1e-5):
        super(DiceLoss, self).__init__()
        self.epsilon = epsilon

    def forward(self, predictions, targets):
        predictions = predictions.view(-1)
        targets = targets.view(-1)
        intersection = (predictions * targets).sum()
        union = predictions.sum() + targets.sum()
        dice_coeff = (2.0 * intersection + self.epsilon) / (union + self.epsilon)
        dice_loss = 1 - dice_coeff
        return dice_loss


class DSCLoss(nn.Module):
    def __init__(
        self, alpha: float = 1.0, smooth: float = 1.0, reduction: str = "mean"
    ):
        super(DSCLoss, self).__init__()
        self.alpha = alpha
        self.smooth = smooth
        self.reduction = reduction

    def forward(self, probs, targets):
        targets = targets.float()
        intersection = (probs * targets).sum(dim=1)
        weighted_intersection = (2.0 * intersection + self.smooth) ** self.alpha

        cardinality = (probs + targets).sum(dim=1)
        weighted_cardinality = (cardinality + self.smooth) ** self.alpha
        dice_coeff = weighted_intersection / weighted_cardinality
        dice_loss = 1 - dice_coeff
        if self.reduction == "mean":
            return dice_loss.mean()
        elif self.reduction == "sum":
            return dice_loss.sum()
        else:
            return dice_loss


class HingeLoss(nn.Module):
    def __init__(self):
        super(HingeLoss, self).__init__()

    def forward(self, y_pred, y_true):

        # convert to range: [-1,1]
        y_true_transformed = 2 * y_true - 1
        y_pred_transformed = 2 * y_pred - 1
        loss = F.relu(1 - y_true_transformed * y_pred_transformed)
        return torch.mean(loss)

def entropy_regularization(logits):
    probs = torch.softmax(logits, dim=1)
    entropy = -torch.sum(
        probs * torch.log(probs.clamp(min=1e-8)), dim=1
    )
    return entropy.mean()


def reference_consistency_loss(a, p, n, cross_triplet=True):
    # Stack the tensors to create a [b, 3] tensor for batch operation
    """
    To minimize the differences in representations of reference () genes across triplet.
    """
    if cross_triplet:
        reference_representations = torch.stack(
            (a.squeeze(), p.squeeze(), n.squeeze()), dim=1
        )
        loss = cv_loss(reference_representations, dim=1)
    else:
        reference_representations = torch.cat([a, p, n], dim=0)
        loss = cv_loss(reference_representations, dim=0)

    return loss


def cv_loss(reference_representations, dim=0):
    """
    To minimize the differences in representations of reference () genes across samples.
    Reasons: ubiquitously expressed genes are refer to genes that are consistently expressed across different samples, tissues, cell types,
    and conditions. Similar to housekeeping genes, ubiquitously expressed genes provide essential cellular functions necessary for the maintenance
    and survival of cells. They are expressed at stable levels, making them useful as reference points or controls in gene expression studies,
    including quantitative PCR and RNA sequencing analyses, to ensure accurate normalization and comparison across samples.

    While absolute expression levels might not match perfectly across all individuals due to biological variability,
    technical factors, and sensitivity of measurement techniques,
    the expression levels of UEGs should be relatively consistent and close to each other among different people.
    This consistency makes UEGs reliable for comparative studies and normalization in gene expression analyses.
    """

    # print(reference_representations)
    mean_ref = torch.mean(reference_representations, dim=dim)
    std_ref = torch.std(reference_representations, dim=dim)
    # 使用平均值的绝对值
    cv = std_ref / (torch.abs(mean_ref) + 1e-6)
    cv_loss = cv.mean()

    return cv_loss


def cv_loss_penalty(reference_representations, dim=0):
    """
    To minimize the differences in representations of reference (UEGs) genes across samples.
    Reasons: ubiquitously expressed genes (UEGs) are consistently expressed across different samples, tissues, cell types, and conditions. Similar to housekeeping genes, UEGs provide essential cellular functions necessary for the maintenance and survival of cells. They are expressed at stable levels, making them useful as reference points or controls in gene expression studies, including quantitative PCR and RNA sequencing analyses, to ensure accurate normalization and comparison across samples.
    While absolute expression levels might not match perfectly across all individuals due to biological variability, technical factors, and sensitivity of measurement techniques, the expression levels of UEGs should be relatively consistent and close to each other among different people.
    This consistency makes UEGs reliable for comparative studies and normalization in gene expression analyses.
    """

    # Ensure the input is positive by adding a penalty for negative values
    penalty = torch.nn.ReLU()(1e-6 - reference_representations).sum()

    # Calculate the mean and standard deviation
    mean_ref = torch.mean(reference_representations, dim=dim)
    std_ref = torch.std(reference_representations, dim=dim)

    # Calculate the coefficient of variation
    cv = std_ref / (torch.abs(mean_ref) + 1e-6)
    cv_loss = cv.mean()

    # Combine cv_loss with the penalty
    total_loss = cv_loss + penalty

    return total_loss


def msd_loss(expression_levels, dim=0, target_mean=1.0, regularization_weight=0.1):
    """
    Minimize variance in expression levels of reference genes across samples, while encouraging expression levels to be close to target_mean.

    Parameters:
    expression_levels (torch.Tensor): Tensor of shape [num_samples, num_genes]
                                      representing expression levels of genes across samples.
    dim (int): Dimension along which to calculate the mean.
    target_mean (float): Target mean expression level.
    regularization_weight (float): Weight of the regularization term.

    Returns:
    torch.Tensor: Scalar loss value.
    """
    # Calculate the mean expression level for each gene across samples
    mean_expression = torch.mean(expression_levels, dim=dim, keepdim=True)

    # Calculate the abs difference from the target mean for each gene
    diffs = torch.abs(expression_levels - target_mean)

    # Calculate the regularization term
    reg_term = torch.mean((expression_levels - target_mean) ** 2)

    # Average the absolute differences across all genes and samples
    loss = torch.mean(diffs) + regularization_weight * reg_term

    return loss


def independence_loss(x, y):
    """
    a loss function that penalizes correlation between x and y.
    """
    y_binary = torch.argmax(y, dim=1).float()
    y_mean = torch.mean(y_binary)
    x_mean = torch.mean(x)
    x_std = torch.std(x, unbiased=True) + 1e-8
    n_pos = torch.sum(y_binary)
    n_neg = len(y_binary) - n_pos
    x_mean_pos = torch.mean(x[y_binary == 1])
    x_mean_neg = torch.mean(x[y_binary == 0])
    pbc = (x_mean_pos - x_mean_neg) * torch.sqrt(n_pos * n_neg) / (x_std * len(x))
    return torch.abs(pbc)


class SupConLoss(nn.Module):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    It also supports the unsupervised contrastive loss in SimCLR"""

    def __init__(self, temperature=0.07, contrast_mode="all", base_temperature=0.07):
        super(SupConLoss, self).__init__()
        self.temperature = temperature
        self.contrast_mode = contrast_mode
        self.base_temperature = base_temperature

    def forward(self, features, labels=None, mask=None):
        """Compute loss for model. If both `labels` and `mask` are None,
        it degenerates to SimCLR unsupervised loss:
        https://arxiv.org/pdf/2002.05709.pdf
        Args:
            features: hidden vector of shape [bsz, n_views, ...].
            labels: ground truth of shape [bsz].
            mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
                has the same class as sample i. Can be asymmetric.
        Returns:
            A loss scalar.
        """
        device = torch.device("cuda") if features.is_cuda else torch.device("cpu")

        if len(features.shape) < 3:
            raise ValueError(
                "`features` needs to be [bsz, n_views, ...],"
                "at least 3 dimensions are required"
            )
        if len(features.shape) > 3:
            features = features.view(features.shape[0], features.shape[1], -1)

        batch_size = features.shape[0]
        if labels is not None and mask is not None:
            raise ValueError("Cannot define both `labels` and `mask`")
        elif labels is None and mask is None:
            mask = torch.eye(batch_size, dtype=torch.float32).to(device)
        elif labels is not None:
            labels = labels.contiguous().view(-1, 1)
            if labels.shape[0] != batch_size:
                raise ValueError("Num of labels does not match num of features")
            mask = torch.eq(labels, labels.T).float().to(device)
        else:
            mask = mask.float().to(device)

        contrast_count = features.shape[1]
        contrast_feature = torch.cat(torch.unbind(features, dim=1), dim=0)
        if self.contrast_mode == "one":
            anchor_feature = features[:, 0]
            anchor_count = 1
        elif self.contrast_mode == "all":
            anchor_feature = contrast_feature
            anchor_count = contrast_count
        else:
            raise ValueError("Unknown mode: {}".format(self.contrast_mode))

        # compute logits
        anchor_dot_contrast = torch.div(
            torch.matmul(anchor_feature, contrast_feature.T), self.temperature
        )
        # for numerical stability
        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        # tile mask
        mask = mask.repeat(anchor_count, contrast_count)
        # mask-out self-contrast cases
        logits_mask = torch.scatter(
            torch.ones_like(mask),
            1,
            torch.arange(batch_size * anchor_count).view(-1, 1).to(device),
            0,
        )
        mask = mask * logits_mask

        # compute log_prob
        exp_logits = torch.exp(logits) * logits_mask
        log_prob = logits - torch.log(exp_logits.sum(1, keepdim=True))

        # compute mean of log-likelihood over positive
        mean_log_prob_pos = (mask * log_prob).sum(1) / mask.sum(1)

        # loss
        loss = -(self.temperature / self.base_temperature) * mean_log_prob_pos
        loss = loss.view(anchor_count, batch_size).mean()

        return loss


def device_as(t1, t2):
    """
    Moves t1 to the device of t2
    """
    return t1.to(t2.device)


class ContrastiveLoss(nn.Module):
    """
    Vanilla Contrastive loss, also called InfoNceLoss as in SimCLR paper
    """

    def __init__(self, batch_size, temperature=0.5):
        super().__init__()
        self.batch_size = batch_size
        self.temperature = temperature
        self.mask = (~torch.eye(batch_size * 2, batch_size * 2, dtype=bool)).float()

    def calc_similarity_batch(self, a, b):
        representations = torch.cat([a, b], dim=0)
        return F.cosine_similarity(
            representations.unsqueeze(1), representations.unsqueeze(0), dim=2
        )

    def forward(self, proj_1, proj_2):
        """
        proj_1 and proj_2 are batched embeddings [batch, embedding_dim]
        where corresponding indices are pairs
        z_i, z_j in the SimCLR paper
        """
        batch_size = proj_1.shape[0]
        z_i = F.normalize(proj_1, p=2, dim=1)
        z_j = F.normalize(proj_2, p=2, dim=1)

        similarity_matrix = self.calc_similarity_batch(z_i, z_j)

        sim_ij = torch.diag(similarity_matrix, batch_size)
        sim_ji = torch.diag(similarity_matrix, -batch_size)

        positives = torch.cat([sim_ij, sim_ji], dim=0)

        nominator = torch.exp(positives / self.temperature)

        denominator = device_as(self.mask, similarity_matrix) * torch.exp(
            similarity_matrix / self.temperature
        )

        all_losses = -torch.log(nominator / torch.sum(denominator, dim=1))
        loss = torch.sum(all_losses) / (2 * self.batch_size)
        return loss



from .scorer import score, score2, score3
from .plotter import plot_embed_with_label, plot_performance
from .sankey import plot_sankey_diagram, get_projector_weights
from .loader import loadcompass
from .crmap import prepare_crmap_data, personal_crmap_data, draw_personal_crmap



from sklearn.metrics import roc_auc_score, precision_recall_curve
from sklearn.metrics import auc as prc_auc_score
from sklearn.metrics import f1_score, accuracy_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import matthews_corrcoef
import numpy as np


def score(y_true, y_prob, y_pred):

    select = ~y_true.isna()
    y_prob = y_prob[select]
    y_true = y_true[select]
    y_pred = y_pred[select]  # .map({'NR':0, 'R':1})

    if len(y_true.unique()) == 1:
        roc = np.nan
        prc = np.nan
    else:
        roc = roc_auc_score(y_true, y_prob)
        _precision, _recall, _ = precision_recall_curve(y_true, y_prob)
        prc = prc_auc_score(_recall, _precision)

    f1 = f1_score(y_true, y_pred, pos_label=1)
    acc = accuracy_score(y_true, y_pred)

    return roc, prc, f1, acc


def score2(y_true, y_prob, y_pred):

    select = ~y_true.isna()
    y_prob = y_prob[select]
    y_true = y_true[select]
    y_pred = y_pred[select]  # .map({'NR':0, 'R':1})

    if len(y_true.unique()) == 1:
        roc = np.nan
        prc = np.nan
    else:
        roc = roc_auc_score(y_true, y_prob)
        _precision, _recall, _ = precision_recall_curve(y_true, y_prob)
        prc = prc_auc_score(_recall, _precision)

    f1 = f1_score(y_true, y_pred, pos_label=1)
    acc = accuracy_score(y_true, y_pred)
    mcc = matthews_corrcoef(y_true, y_pred)

    return roc, prc, f1, acc, mcc


def score3(y_true, y_prob, y_pred):
    """
    Calculate evaluation metrics including ROC-AUC, PRC-AUC, F1, Accuracy, MCC, FPR, and FNR.

    Parameters:
    y_true (array-like): True binary labels.
    y_prob (array-like): Predicted probabilities for the positive class.
    y_pred (array-like): Predicted binary labels.

    Returns:
    tuple: roc, prc, f1, acc, mcc, fpr, fnr
    """
    # Filter non-NA values
    select = ~y_true.isna()
    y_prob = y_prob[select]
    y_true = y_true[select]
    y_pred = y_pred[select]  # .map({'NR':0, 'R':1})

    # Initialize ROC-AUC and PRC-AUC
    if len(y_true.unique()) == 1:
        roc = np.nan
        prc = np.nan
    else:
        roc = roc_auc_score(y_true, y_prob)
        _precision, _recall, _ = precision_recall_curve(y_true, y_prob)
        prc = prc_auc_score(_recall, _precision)

    # Calculate F1, Accuracy, and MCC
    f1 = f1_score(y_true, y_pred, pos_label=1)
    acc = accuracy_score(y_true, y_pred)
    mcc = matthews_corrcoef(y_true, y_pred)

    # Calculate Confusion Matrix and derive FPR and FNR
    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
    fpr = fp / (fp + tn) if (fp + tn) > 0 else np.nan  # False Positive Rate
    fnr = fn / (fn + tp) if (fn + tp) > 0 else np.nan  # False Negative Rate

    return roc, prc, f1, acc, mcc, fpr, fnr



from .model import Compass
from .loss import TripletLoss, TriSimplexLoss
from .loss import MAEWithNaNLabelsLoss, CEWithNaNLabelsLoss, FocalLoss
from .loss import DiceLoss, DSCLoss, HingeLoss
from .saver import SaveBestModel
from .scaler import NoScaler, P2Normalizer, Datascaler
from .train import PT_Trainer, PT_Tester
from .tune import FT_Trainer, FT_Tester
from .tune import Predictor, Evaluator, Extractor, Projector
from .adapt import Adp_Trainer, Adp_Tester


from .scorer import score, score2

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from umap import UMAP
import seaborn as sns

sns.set(style="white", font_scale=1.5)


def plot_embed_with_label(
    dfp,
    label_col=["cancer_type"],
    label_type=["c"],
    orders=[None],
    figsize=(10, 10),
    metric="correlation",
    spread=1,
    n_neighbors=5,
    min_dist=0.5,
    s=5,
    random_state=123,
    verbose=False,
    return_coord=False,
    cmap="bright",
    **kwargs,
):
    """
    dfp: dataframe of samples x genes, with a column contains batch information
    label_col: list of labels to be used
    label_type: list of the label types, 'c' for categorical label, 'r' for continous label
    """

    glist = dfp.columns[~dfp.columns.isin(label_col)]
    if len(glist) == 2:
        df2d = dfp[glist]
    else:
        mp = UMAP(
            spread=spread,
            min_dist=min_dist,
            n_neighbors=n_neighbors,
            metric=metric,
            random_state=random_state,
            verbose=verbose,
            **kwargs,
        )  # , metric='correlation'
        embed = mp.fit_transform(dfp[glist])
        df2d = pd.DataFrame(embed, columns=["UMAP1", "UMAP2"], index=dfp.index)
    col1, col2 = df2d.columns
    df2d = df2d.join(dfp[label_col])

    figs = []
    for label, t, order in zip(label_col, label_type, orders):
        fig, ax = plt.subplots(figsize=figsize)
        if t == "c":
            cohort_order = order  # df2d.groupby(label).size().sort_values().index
            if cohort_order is None:
                cohort_order = df2d.groupby(label).size().sort_values().index
            colors = sns.color_palette(cmap, len(cohort_order)).as_hex()

            for bt, c in zip(cohort_order, colors):
                dfp1 = df2d[df2d[label] == bt]
                ax.scatter(dfp1[col1], dfp1[col2], label=bt, s=s, c=c)
                # print(color)
            if len(cohort_order) <= 16:
                ax.legend(
                    loc="center left",
                    ncol=1,
                    prop={"size": 12},
                    bbox_to_anchor=(1, 0.5),
                )  #
            else:
                ax.legend(
                    loc="center left",
                    ncol=3,
                    prop={"size": 12},
                    bbox_to_anchor=(1, 0.5),
                )  #
        else:
            ax.scatter(
                df2d[col1], df2d[col2], label=label, s=s, c=df2d[label], cmap=cmap
            )
            ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

        ax.tick_params(
            bottom="on",
            left="off",
            labelleft="on",
            labelbottom="on",
            pad=-0.6,
        )
        sns.despine(top=True, right=True, left=False, bottom=False)
        ax.set_xlabel(col1)
        ax.set_ylabel(col2)
        ax.set_title(label)

        figs.append(fig)

    if not return_coord:
        return figs

    else:
        return figs, df2d


def plot_performance(y_true, y_prob, y_pred):

    from sklearn.metrics import confusion_matrix

    roc, prc, f1, acc, mcc = score2(y_true, y_prob, y_pred)
    dfp = pd.DataFrame([y_true, y_prob, y_pred]).T
    dfp.columns = ["Label", "Pred. Prob.", "Pred_label"]
    dfp.Label = dfp.Label.map({0: "NR", 1: "R"})
    cf_matrix = confusion_matrix(y_true, y_pred, labels=[1, 0])

    tp_and_fn = cf_matrix.sum(1)
    tp_and_fp = cf_matrix.sum(0)
    tp = cf_matrix.diagonal()
    precision = tp / tp_and_fp
    recall = tp / tp_and_fn
    precision, recall = precision[0], recall[0]

    palette = sns.color_palette("rainbow", 12)

    colors = palette.as_hex()
    boxpalette = {"NR": colors[1], "R": colors[-3]}
    swarmpalette = {"NR": colors[2], "R": colors[-3]}

    fig, axes = plt.subplots(
        ncols=3,
        nrows=1,
        figsize=(8, 3),
        gridspec_kw={"width_ratios": [4, 1, 3]},
        sharex=False,
        sharey=False,
    )

    ax2, ax1, ax3 = axes

    ###################################
    order = ["R", "NR"]
    sns.boxplot(
        data=dfp,
        x="Label",
        y="Pred. Prob.",
        fliersize=0.0,
        width=0.5,
        order=order,
        ax=ax1,
        palette=boxpalette,
        saturation=0.8,
        boxprops={"facecolor": "None"},
    )
    sns.stripplot(
        data=dfp,
        x="Label",
        y="Pred. Prob.",
        ax=ax1,
        size=3,
        order=order,
        palette=boxpalette,
        edgecolor="k",
        linewidth=0.1,
    )

    ax1.xaxis.tick_bottom()  # x axis on top
    ax1.xaxis.set_label_position("bottom")
    ax1.set_xlabel("")
    ax1.tick_params(axis="x", labelrotation=60)

    ax1.set_yticks([0.0, 0.5, 1.0])
    ax1.yaxis.tick_left()  # x axis on top
    ax1.spines[["right", "top"]].set_visible(False)

    ###################################
    group_names = ["True R", "False NR", "False R", "True NR"]
    group_counts = ["{0:0.0f}".format(value) for value in cf_matrix.flatten()]
    group_percentages = [
        "{0:.2%}".format(value) for value in cf_matrix.flatten() / np.sum(cf_matrix)
    ]
    labels = [
        f"{v1}\n{v2}" for v1, v2 in zip(group_names, group_counts)
    ]  # ,group_percentages
    labels = np.asarray(labels).reshape(2, 2)

    cf_df = pd.DataFrame(cf_matrix, index=["R", "NR"], columns=["R", "NR"])
    sns.heatmap(cf_df, annot=labels, fmt="", cmap="Blues", ax=ax2, cbar=False)

    ax2.xaxis.tick_bottom()  # x axis on top
    ax2.xaxis.set_label_position("bottom")
    # ax2.set_xlabel("Predicted Label")
    ax2.tick_params(axis="x", labelrotation=60)

    ax2.yaxis.tick_left()  # x axis on top
    ax2.yaxis.set_label_position("left")
    ax2.set_ylabel("True Label")
    ax2.tick_params(axis="y", labelrotation=60)

    ###################################
    dfpp = pd.DataFrame(
        [mcc, roc, prc, f1, acc], index=["MCC", "ROC", "PRC", "F1", "ACC"]
    )

    dfpp.plot(kind="barh", ax=ax3, legend=False, color="b", alpha=0.5)
    ax3.yaxis.tick_left()  # x axis on top
    ax3.yaxis.set_label_position("left")
    ax3.set_xticks([0.0, 0.5, 1.0])
    ax3.xaxis.tick_bottom()  # x axis on top
    ax3.xaxis.set_label_position("bottom")

    for y, x in dfpp[0].reset_index(drop=True).items():
        ax3.text(x, y - 0.15, "%.2f" % x)
    ax3.spines[["right", "top"]].set_visible(False)

    fig.tight_layout(pad=1.5)
    return fig



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  3 13:31:25 2023

@author: Wanxiang Shen

"""
import torch
import torch.nn as nn
from torch import Tensor
import enum
import math
import os


cwd = os.path.dirname(__file__)


class _GeneInitialization(enum.Enum):

    UNIFORM = "uniform"
    NORMAL = "normal"

    @classmethod
    def from_str(cls, initialization: str) -> "_GeneInitialization":
        try:
            return cls(initialization)
        except ValueError:
            valid_values = [x.value for x in _GeneInitialization]
            raise ValueError(f"initialization must be one of {valid_values}")

    def apply(self, x: Tensor, d: int) -> None:
        d_sqrt_inv = 1 / math.sqrt(d)
        if self == _GeneInitialization.UNIFORM:
            nn.init.uniform_(x, a=-d_sqrt_inv, b=d_sqrt_inv)
        elif self == _GeneInitialization.NORMAL:
            nn.init.normal_(x, std=d_sqrt_inv)


class _GeneExpressionEmbedding(nn.Module):
    """Embeds the gene expression data.

    For one gene, the embeddings consists of two steps:

    * the input abundence features is multiplied by a trainable vector (gene expression embedding)
    * another trainable positinal embedding vector is added (gene postional encoding)

    Note that each gene has its separate pair of trainable embeddings, i.e. the embedding vectors
    are not shared between genes.

    """

    def __init__(
        self,
        n_features: int,
        d_gene: int,
        bias: bool,
        initialization: str,
    ) -> None:
        """
        Args:
            n_features: the number of genes
            d_gene: the embed size of one gene
            bias: if `False`, then the transformation will include only multiplication.
                **Warning**: :code:`bias=False` means no positional encoding (PE).
            initialization: initialization policy for parameters. Must be one of
                :code:`['uniform', 'normal']`.

        References:
            *
            https://yura52.github.io/rtdl/stable/_modules/rtdl/modules.html#NumericalFeatureTokenizer

        """
        super().__init__()
        initialization_ = _GeneInitialization.from_str(initialization)
        self.weight = nn.Parameter(Tensor(n_features, d_gene))
        self.bias = nn.Parameter(Tensor(n_features, d_gene)) if bias else None
        for parameter in [self.weight, self.bias]:
            if parameter is not None:
                initialization_.apply(parameter, d_gene)

    @property
    def n_genes(self) -> int:
        """The number of genes."""
        return len(self.weight)

    @property
    def d_gene(self) -> int:
        """The size of one gene."""
        return self.weight.shape[1]

    def forward(self, x: Tensor) -> Tensor:
        x = self.weight[None] * x[..., None]
        if self.bias is not None:
            x = x + self.bias[None]
        return x


class AbundanceEmbedding(nn.Module):
    def __init__(self, n_genes: int, embed_dim: int, learnable_pe=True) -> None:
        """
        n_genes:  Number of genes
        embed_dim: gene embedding size
        learnable_pe: {True, False}
        """
        super().__init__()
        gee_layer = _GeneExpressionEmbedding(
            n_genes, embed_dim, learnable_pe, "uniform"
        )
        layers = [gee_layer, nn.ReLU()]
        # layers = [gee_layer]
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)


class PositionalEncoding(nn.Module):
    def __init__(self, input_dim=876, dim=32, type="gene2vect"):
        """
        type:{'gene2vect', 'umap', 'pumap'}
        """
        super().__init__()
        pefile = os.path.join(cwd, "pe/%s_%s_%s.pt" % (type, input_dim, dim))
        pe = torch.load(pefile)  # 876,32
        pe = pe.unsqueeze(0)  # 1,876,32
        self.register_buffer("pe", pe)

    def forward(self, x):
        pe = self.pe[:, : x.size(1)].requires_grad_(False)
        return pe


class GeneEmbedding(nn.Module):
    def __init__(self, input_dim=876, d_model=32, pos_emb=None):
        """
        pos_emb: {None, 'learnable', 'gene2vect', 'umap'}
        """
        super(GeneEmbedding, self).__init__()

        self.input_dim = input_dim
        self.pos_emb = pos_emb
        self.d_model = d_model

        if pos_emb == "learnable":
            self.abundance_embedder = AbundanceEmbedding(
                input_dim, d_model, learnable_pe=True
            )
            self.positional_encoder = torch.zeros_like
        else:
            self.abundance_embedder = AbundanceEmbedding(
                input_dim, d_model, learnable_pe=False
            )
            if pos_emb is None:
                self.positional_encoder = torch.zeros_like
            else:
                self.positional_encoder = PositionalEncoding(
                    input_dim, d_model, pos_emb
                )

    def forward(self, x):
        """
        Args:
          x: embeddings (batch_size, seq_length)

        Returns:
            gene value embeddings (batch_size, seq_length, d_model) + gene positional embedding (batch_size, seq_length, d_model)
        """
        if (self.pos_emb is None) | (self.pos_emb == "learnable"):
            x = self.abundance_embedder(x)
        else:
            x = self.abundance_embedder(x)  # blc
            pe = self.positional_encoder(x)  # blc
            x = x + pe
        return x



import torch
import torch.nn.functional as F
from functools import partial
import numpy as np

from torch import Tensor
from typing import Optional
from torch import nn, einsum
import math

from einops import rearrange, repeat, pack, unpack

from .norm import create_norm
from .transformer import AbstractTrasnformerLayer


TOKEN_SELF_ATTN_VALUE = -5e4


def exists(val):
    return val is not None


def empty(tensor):
    return tensor.numel() == 0


def default(val, d):
    return val if exists(val) else d


def to(t):
    return {"device": t.device, "dtype": t.dtype}


def max_neg_value(tensor):
    return -torch.finfo(tensor.dtype).max


def l2norm(tensor):
    dtype = tensor.dtype
    normed = F.normalize(tensor, dim=-1)
    return normed.type(dtype)


def pad_to_multiple(tensor, multiple, dim=-1, value=0):
    seqlen = tensor.shape[dim]
    m = seqlen / multiple
    if m.is_integer():
        return False, tensor
    remainder = math.ceil(m) * multiple - seqlen
    pad_offset = (0,) * (-1 - dim) * 2
    return True, F.pad(tensor, (*pad_offset, 0, remainder), value=value)


def look_around(x, backward=1, forward=0, pad_value=-1, dim=2):
    t = x.shape[1]
    dims = (len(x.shape) - dim) * (0, 0)
    padded_x = F.pad(x, (*dims, backward, forward), value=pad_value)
    tensors = [
        padded_x[:, ind : (ind + t), ...] for ind in range(forward + backward + 1)
    ]
    return torch.cat(tensors, dim=dim)


class SinusoidalEmbeddings(nn.Module):
    def __init__(self, dim):
        super().__init__()
        inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

    def forward(self, x):
        n = x.shape[-2]
        t = torch.arange(n, device=x.device).type_as(self.inv_freq)
        freqs = torch.einsum("i , j -> i j", t, self.inv_freq)
        return torch.cat((freqs, freqs), dim=-1)


def rotate_half(x):
    x = rearrange(x, "b ... (r d) -> b (...) r d", r=2)
    x1, x2 = x.unbind(dim=-2)
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_pos_emb(q, k, freqs):
    q, k = map(lambda t: (t * freqs.cos()) + (rotate_half(t) * freqs.sin()), (q, k))
    return q, k


def orthogonal_matrix_chunk(cols, device=None):
    unstructured_block = torch.randn((cols, cols), device=device)
    # q, r = torch.qr(unstructured_block.cpu(), some = True)
    q, r = torch.linalg.qr(unstructured_block.cpu(), "reduced")
    q, r = map(lambda t: t.to(device), (q, r))
    return q.t()


def softmax_kernel(
    data, *, projection_matrix, is_query, normalize_data=True, eps=1e-4, device=None
):
    b, h, *_ = data.shape

    data_normalizer = (data.shape[-1] ** -0.25) if normalize_data else 1.0

    ratio = projection_matrix.shape[0] ** -0.5

    projection = repeat(projection_matrix, "j d -> b h j d", b=b, h=h)
    projection = projection.type_as(data)

    data_dash = torch.einsum("...id,...jd->...ij", (data_normalizer * data), projection)

    diag_data = data**2
    diag_data = torch.sum(diag_data, dim=-1)
    diag_data = (diag_data / 2.0) * (data_normalizer**2)
    diag_data = diag_data.unsqueeze(dim=-1)

    if is_query:
        data_dash = ratio * (
            torch.exp(
                data_dash
                - diag_data
                - torch.max(data_dash, dim=-1, keepdim=True).values
            )
            + eps
        )
    else:
        data_dash = ratio * (
            torch.exp(data_dash - diag_data - torch.max(data_dash)) + eps
        )

    return data_dash.type_as(data)


def generalized_kernel(
    data,
    *,
    projection_matrix,
    kernel_fn=nn.ReLU(),
    kernel_epsilon=0.001,
    normalize_data=True,
    device=None,
):
    b, h, *_ = data.shape

    data_normalizer = (data.shape[-1] ** -0.25) if normalize_data else 1.0

    if projection_matrix is None:
        return kernel_fn(data_normalizer * data) + kernel_epsilon

    projection = repeat(projection_matrix, "j d -> b h j d", b=b, h=h)
    projection = projection.type_as(data)

    data_dash = torch.einsum("...id,...jd->...ij", (data_normalizer * data), projection)

    data_prime = kernel_fn(data_dash) + kernel_epsilon
    return data_prime.type_as(data)


# non-causal linear attention
def linear_attention(q, k, v):
    k_cumsum = k.sum(dim=-2)
    D_inv = 1.0 / torch.einsum("...nd,...d->...n", q, k_cumsum.type_as(q))
    context = torch.einsum("...nd,...ne->...de", k, v)
    out = torch.einsum("...de,...nd,...n->...ne", context, q, D_inv)
    return out


# efficient causal linear attention, created by EPFL
def causal_linear_attention(q, k, v, eps=1e-6):
    raise NotImplementedError("Please refer to performer-pytorch repo!")
    # https://github.com/lucidrains/performer-pytorch/blob/main/performer_pytorch/performer_pytorch.py


def causal_linear_attention_noncuda(q, k, v, chunk_size=128, eps=1e-6):
    raise NotImplementedError("Please refer to performer-pytorch repo!")
    # https://github.com/lucidrains/performer-pytorch/blob/main/performer_pytorch/performer_pytorch.py


def gaussian_orthogonal_random_matrix(nb_rows, nb_columns, scaling=0, device=None):
    nb_full_blocks = int(nb_rows / nb_columns)

    block_list = []

    for _ in range(nb_full_blocks):
        q = orthogonal_matrix_chunk(nb_columns, device=device)
        block_list.append(q)

    remaining_rows = nb_rows - nb_full_blocks * nb_columns
    if remaining_rows > 0:
        q = orthogonal_matrix_chunk(nb_columns, device=device)
        block_list.append(q[:remaining_rows])

    final_matrix = torch.cat(block_list)

    if scaling == 0:
        multiplier = torch.randn((nb_rows, nb_columns), device=device).norm(dim=1)
    elif scaling == 1:
        multiplier = math.sqrt((float(nb_columns))) * torch.ones(
            (nb_rows,), device=device
        )
    else:
        raise ValueError(f"Invalid scaling {scaling}")

    return torch.diag(multiplier) @ final_matrix


class LocalAttention(nn.Module):
    def __init__(
        self,
        window_size,
        causal=False,
        look_backward=1,
        look_forward=None,
        dropout=0.0,
        shared_qk=False,
        rel_pos_emb_config=None,
        dim=None,
        autopad=False,
        exact_windowsize=False,
        scale=None,
    ):
        super().__init__()
        look_forward = default(look_forward, 0 if causal else 1)
        assert not (causal and look_forward > 0), "you cannot look forward if causal"

        self.scale = scale

        self.window_size = window_size
        self.autopad = autopad
        self.exact_windowsize = exact_windowsize

        self.causal = causal

        self.look_backward = look_backward
        self.look_forward = look_forward

        self.dropout = nn.Dropout(dropout)

        self.shared_qk = shared_qk

        # relative positions

        self.rel_pos = None
        if exists(rel_pos_emb_config) or exists(
            dim
        ):  # backwards compatible with old `rel_pos_emb_config` deprecated argument
            if exists(rel_pos_emb_config):
                dim = rel_pos_emb_config[0]
            self.rel_pos = SinusoidalEmbeddings(dim)

    def forward(self, q, k, v, mask=None, input_mask=None):
        mask = default(mask, input_mask)

        (
            shape,
            autopad,
            pad_value,
            window_size,
            causal,
            look_backward,
            look_forward,
            shared_qk,
        ) = (
            q.shape,
            self.autopad,
            -1,
            self.window_size,
            self.causal,
            self.look_backward,
            self.look_forward,
            self.shared_qk,
        )

        # https://github.com/arogozhnikov/einops/blob/master/docs/4-pack-and-unpack.ipynb
        (q, packed_shape), (k, _), (v, _) = map(lambda t: pack([t], "* n d"), (q, k, v))

        # rotary embeddings

        if exists(self.rel_pos):
            pos_emb = self.rel_pos(q)
            q, k = apply_rotary_pos_emb(q, k, pos_emb)

        # auto padding

        if autopad:
            orig_seq_len = q.shape[1]
            (needed_pad, q), (_, k), (_, v) = map(
                lambda t: pad_to_multiple(t, self.window_size, dim=-2), (q, k, v)
            )

        b, n, dim_head, device, dtype = *q.shape, q.device, q.dtype

        scale = default(self.scale, dim_head**-0.5)

        assert (
            n % window_size
        ) == 0, f"sequence length {n} must be divisible by window size {window_size} for local attention"

        windows = n // window_size

        if shared_qk:
            k = l2norm(k)

        seq = torch.arange(n, device=device)
        b_t = rearrange(seq, "(w n) -> 1 w n", w=windows, n=window_size)

        bq, bk, bv = map(
            lambda t: rearrange(t, "b (w n) d -> b w n d", w=windows), (q, k, v)
        )

        look_around_kwargs = dict(
            backward=look_backward, forward=look_forward, pad_value=pad_value
        )

        bk = look_around(bk, **look_around_kwargs)
        bv = look_around(bv, **look_around_kwargs)

        bq_t = b_t
        bq_k = look_around(b_t, **look_around_kwargs)

        bq_t = rearrange(bq_t, "... i -> ... i 1")
        bq_k = rearrange(bq_k, "... j -> ... 1 j")

        sim = einsum("b h i e, b h j e -> b h i j", bq, bk) * scale

        mask_value = max_neg_value(sim)

        if shared_qk:
            self_mask = bq_t == bq_k
            sim = sim.masked_fill(self_mask, TOKEN_SELF_ATTN_VALUE)
            del self_mask

        if causal:
            causal_mask = bq_t < bq_k

            if self.exact_windowsize:
                max_causal_window_size = self.window_size * self.look_backward
                causal_mask = causal_mask | (bq_t > (bq_k + max_causal_window_size))

            sim = sim.masked_fill(causal_mask, mask_value)
            del causal_mask

        # mask out padding value

        if autopad and needed_pad:
            pad_mask = bq_k == pad_value
            sim = sim.masked_fill(pad_mask, mask_value)
            del pad_mask

        if exists(mask):
            batch = mask.shape[0]
            assert (b % batch) == 0

            h = b // mask.shape[0]

            if autopad:
                _, mask = pad_to_multiple(mask, window_size, dim=-1, value=False)

            mask = rearrange(mask, "... (w n) -> (...) w n", w=windows, n=window_size)
            mask = look_around(mask, **{**look_around_kwargs, "pad_value": False})
            mask = rearrange(mask, "... j -> ... 1 j")
            mask = repeat(mask, "b ... -> (b h) ...", h=h)
            sim = sim.masked_fill(~mask, mask_value)
            del mask

        # attention

        attn = sim.softmax(dim=-1)
        attn = self.dropout(attn)

        # aggregation

        out = einsum("b h i j, b h j e -> b h i e", attn, bv)
        out = rearrange(out, "b w n d -> b (w n) d")

        if autopad:
            out = out[:, :orig_seq_len, :]

        out, *_ = unpack(out, packed_shape, "* n d")
        return out


class FastAttention(nn.Module):
    def __init__(
        self,
        dim_heads,
        nb_features=None,
        ortho_scaling=0,
        causal=False,
        generalized_attention=False,
        kernel_fn=nn.ReLU(),
        no_projection=False,
    ):
        super().__init__()
        nb_features = default(nb_features, int(dim_heads * math.log(dim_heads)))

        self.dim_heads = dim_heads
        self.nb_features = nb_features
        self.ortho_scaling = ortho_scaling

        self.create_projection = partial(
            gaussian_orthogonal_random_matrix,
            nb_rows=self.nb_features,
            nb_columns=dim_heads,
            scaling=ortho_scaling,
        )
        projection_matrix = self.create_projection()
        self.register_buffer("projection_matrix", projection_matrix)

        self.generalized_attention = generalized_attention
        self.kernel_fn = kernel_fn

        # if this is turned on, no projection will be used
        # queries and keys will be softmax-ed as in the original efficient attention paper
        self.no_projection = no_projection

        self.causal = causal
        if causal:
            try:
                import fast_transformers.causal_product.causal_product_cuda

                self.causal_linear_fn = partial(causal_linear_attention)
            except ImportError:
                print(
                    "unable to import cuda code for auto-regressive Performer. will default to the memory inefficient non-cuda version"
                )
                self.causal_linear_fn = causal_linear_attention_noncuda

    @torch.no_grad()
    def redraw_projection_matrix(self, device):
        projections = self.create_projection(device=device)
        self.projection_matrix.copy_(projections)
        del projections

    def forward(self, q, k, v, output_attentions=False):
        device = q.device
        # inds = [8060, 8064, 6243, 8575, 10342, 10913, 9366, 993, 7796, 5210, 5212, 5504, 6851, 6559, 5508, 13107, 13820]
        if self.no_projection:
            q = q.softmax(dim=-1)
            k = torch.exp(k) if self.causal else k.softmax(dim=-2)

        elif self.generalized_attention:
            create_kernel = partial(
                generalized_kernel,
                kernel_fn=self.kernel_fn,
                projection_matrix=self.projection_matrix,
                device=device,
            )
            q, k = map(create_kernel, (q, k))

        else:
            create_kernel = partial(
                softmax_kernel, projection_matrix=self.projection_matrix, device=device
            )
            q = create_kernel(q, is_query=True)
            k = create_kernel(k, is_query=False)

        attn_fn = linear_attention if not self.causal else self.causal_linear_fn
        out = attn_fn(q, k, v)
        if output_attentions:
            v_diag = torch.eye(v.shape[-2]).to(device)
            v_diag = (
                v_diag.unsqueeze(0).unsqueeze(0).repeat(v.shape[0], v.shape[1], 1, 1)
            )
            # attn_weights = torch.zeros(1, 1, len(inds), len(inds)).to(device).to(torch.float16)
            # attn_weights = torch.zeros(1, q.shape[1], len(inds), len(inds)).to(device).to(torch.float16)
            attn_weights = (
                torch.zeros(1, 1, q.shape[2], q.shape[2]).to(device).to(torch.float16)
            )
            for head_dim in range(q.shape[1]):
                # attn_weights[0, head_dim] = torch.abs(attn_fn(q[:,head_dim].to(torch.float16), k[:,head_dim].to(torch.float16), v_diag[:,head_dim].to(torch.float16)))[0, inds][:, inds]
                attn_weights += torch.abs(
                    attn_fn(
                        q[:, head_dim].to(torch.float16),
                        k[:, head_dim].to(torch.float16),
                        v_diag[:, head_dim].to(torch.float16),
                    )
                )
                # attn_weights += norm_tensor(torch.abs(attn_fn(q[:,head_dim].to(torch.float16), k[:,head_dim].to(torch.float16), v_diag[:,head_dim].to(torch.float16))), dim=-1)
            attn_weights /= q.shape[1]
            return out, attn_weights
        else:
            return out


class PerformerAttention(nn.Module):
    def __init__(
        self,
        dim,
        causal=False,
        heads=8,
        dim_head=32,
        local_heads=0,
        local_window_size=256,
        nb_features=None,
        feature_redraw_interval=1000,
        generalized_attention=False,
        kernel_fn=nn.ReLU(),
        dropout=0.0,
        no_projection=False,
        qkv_bias=False,
    ):
        super().__init__()
        assert dim % heads == 0, "dimension must be divisible by number of heads"
        dim_head = default(dim_head, dim // heads)
        inner_dim = dim_head * heads
        self.fast_attention = FastAttention(
            dim_head,
            nb_features,
            causal=causal,
            generalized_attention=generalized_attention,
            kernel_fn=kernel_fn,
            no_projection=no_projection,
        )

        self.heads = heads
        self.global_heads = heads - local_heads
        self.local_attn = (
            LocalAttention(
                window_size=local_window_size,
                causal=causal,
                autopad=True,
                dropout=dropout,
                look_forward=int(not causal),
                rel_pos_emb_config=(dim_head, local_heads),
            )
            if local_heads > 0
            else None
        )

        self.to_q = nn.Linear(dim, inner_dim, bias=qkv_bias)
        self.to_k = nn.Linear(dim, inner_dim, bias=qkv_bias)
        self.to_v = nn.Linear(dim, inner_dim, bias=qkv_bias)
        self.to_out = nn.Linear(inner_dim, dim)
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        x,
        pos_emb=None,
        context=None,
        mask=None,
        context_mask=None,
        output_attentions=False,
        **kwargs,
    ):
        b, n, _, h, gh = *x.shape, self.heads, self.global_heads

        cross_attend = exists(context)

        context = default(context, x)
        context_mask = default(context_mask, mask) if not cross_attend else context_mask

        q, k, v = self.to_q(x), self.to_k(context), self.to_v(context)

        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))
        (q, lq), (k, lk), (v, lv) = map(lambda t: (t[:, :gh], t[:, gh:]), (q, k, v))

        attn_outs = []

        if not empty(q):
            if exists(context_mask):
                global_mask = context_mask[:, None, :, None]
                v.masked_fill_(~global_mask, 0.0)

            if exists(pos_emb) and not cross_attend:
                (
                    q,
                    k,
                ) = apply_rotary_pos_emb(q, k, pos_emb)

            if output_attentions:
                out, attn_weights = self.fast_attention(q, k, v, output_attentions)
            else:
                out = self.fast_attention(q, k, v)
            attn_outs.append(out)

        if not empty(lq):
            assert (
                not cross_attend
            ), "local attention is not compatible with cross attention"
            out = self.local_attn(lq, lk, lv, input_mask=mask)
            attn_outs.append(out)

        out = torch.cat(
            attn_outs, dim=1
        )  # combine attn_out and cross_attn_out, here we have only attn_out, that means this line does nothing
        out = rearrange(out, "b h n d -> b n (h d)")
        out = self.to_out(out)
        if output_attentions:
            return self.dropout(out), attn_weights
        else:
            return self.dropout(out), None


class PerformerLayer(nn.Module, AbstractTrasnformerLayer):
    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        norm="layernorm",
        norm_first: bool = False,
        causal=False,
    ):
        super().__init__()
        self.self_attn = PerformerAttention(
            dim=embed_dim, heads=num_heads, dropout=dropout
        )
        self._ff_block = nn.Sequential(
            nn.Linear(embed_dim, embed_dim * 2),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(embed_dim * 2, embed_dim),
            nn.Dropout(dropout),
        )

        self.dropout1 = nn.Dropout(dropout)
        self.norm = norm
        self.norm1 = create_norm(norm, embed_dim)
        self.norm2 = create_norm(norm, embed_dim)
        self.norm_first = norm_first
        self.support_output_attentions = True

    def _sa_block(self, x: Tensor, output_attentions=False):
        # x = x.unsqueeze(0)
        out, attn_weights = self.self_attn(x, output_attentions=output_attentions)
        return out, attn_weights

    def forward(self, x, output_attentions=False):
        # print(x.shape)
        if self.norm_first:
            x_prime, attn = self._sa_block(
                x=self.norm1(x), output_attentions=output_attentions
            )
            x = x + x_prime
            x = x + self._ff_block(self.norm2(x))
        else:
            x_prime, attn = self._sa_block(x, output_attentions=output_attentions)
            x = self.norm1(x + x_prime)
            x = self.norm2(x + self._ff_block(x))

        return x, attn



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 16:44:09 2023

@author: Wanxiang Shen
"""
import torch
import torch.utils.data as Torchdata
import pandas as pd
import numpy as np
from tqdm import tqdm

tqdm.pandas(ascii=True)

from ..dataloader import GeneData
from .loss import entropy_regularization, independence_loss
from .loss import reference_consistency_loss


def worker_init_fn(worker_id):
    seed = torch.initial_seed() % 2**32
    np.random.seed(seed)


def FT_Trainer(
    train_loader,
    model,
    optimizer,
    ssl_loss,
    tsk_loss,
    device,
    alpha=0.0,
    correction=0.0,
    entropy_weight=0.0,
):

    model.train()
    total_loss = []
    total_ssl_loss = []
    total_tsk_loss = []

    # torch.autograd.set_detect_anomaly(True)
    # for data in tqdm(train_loader, ascii=True):
    for data in train_loader:

        triplet, label = data

        anchor_y_true, positive_y_true, negative_y_true = label
        anchor, positive, negative = triplet

        anchor, positive, negative = (
            anchor.to(device),
            positive.to(device),
            negative.to(device),
        )
        anchor_y_true = anchor_y_true.to(device)
        positive_y_true = positive_y_true.to(device)
        negative_y_true = negative_y_true.to(device)

        optimizer.zero_grad()

        (anchor_emb, anchor_refg), anchor_y_pred = model(anchor)
        (positive_emb, positive_refg), positive_y_pred = model(positive)
        (negative_emb, negative_refg), negative_y_pred = model(negative)

        lss = ssl_loss(anchor_emb, positive_emb, negative_emb)

        ## remove batch effects by minimal the differences between house-keeping genes
        if correction != 0:
            ref = reference_consistency_loss(
                anchor_refg[1], positive_refg[1], negative_refg[1]
            )
            lss = (1 - correction) * lss + correction * ref
            # print("Ref: {:.6f} - lss: {:.2f}".format(ref.item(), lss.item()))

        y_pred = anchor_y_pred  # torch.cat([anchor_y_pred, positive_y_pred, negative_y_pred])
        y_true = anchor_y_true  # torch.cat([anchor_y_true, positive_y_true, negative_y_true])
        tsk = tsk_loss(y_pred, y_true)

        if entropy_weight != 0:
            entropy_reg = entropy_regularization(y_pred)
            tsk = tsk * (1 - entropy_weight) + entropy_reg * entropy_weight

        loss = (1.0 - alpha) * lss + tsk * alpha

        # print(cv_loss, lss, loss)
        loss.backward()
        optimizer.step()

        total_loss.append(loss.item())
        total_ssl_loss.append(lss.item())
        total_tsk_loss.append(tsk.item())

    train_total_loss = np.mean(total_loss)
    train_ssl_loss = np.mean(total_ssl_loss)
    train_tsk_loss = np.mean(total_tsk_loss)

    return train_total_loss, train_ssl_loss, train_tsk_loss


@torch.no_grad()
def FT_Tester(test_loader, model, ssl_loss, tsk_loss, device, alpha=0.0, correction=0):
    model.eval()
    total_loss = []
    total_ssl_loss = []
    total_tsk_loss = []

    for data in test_loader:
        triplet, label = data
        anchor, positive, negative = triplet
        anchor_y_true, positive_y_true, negative_y_true = label

        anchor = anchor.to(device)
        positive = positive.to(device)
        negative = negative.to(device)
        anchor_y_true = anchor_y_true.to(device)
        positive_y_true = positive_y_true.to(device)
        negative_y_true = negative_y_true.to(device)

        (anchor_emb, anchor_refg), anchor_y_pred = model(anchor)

        (positive_emb, positive_refg), positive_y_pred = model(positive)
        (negative_emb, negative_refg), negative_y_pred = model(negative)

        lss = ssl_loss(anchor_emb, positive_emb, negative_emb)

        if correction != 0:
            ref = reference_consistency_loss(
                anchor_refg[1], positive_refg[1], negative_refg[1]
            )
            lss = (1 - correction) * lss + correction * ref
            # print("Ref: {:.6f} - lss: {:.2f}".format(ref.item(), lss.item()))

        # torch.cat([anchor_y_pred, positive_y_pred, negative_y_pred], dim = 0)
        y_pred = anchor_y_pred
        y_true = anchor_y_true
        tsk = tsk_loss(y_pred, y_true)

        loss = (1.0 - alpha) * lss + tsk * alpha

        total_loss.append(loss.item())
        total_ssl_loss.append(lss.item())
        total_tsk_loss.append(tsk.item())

    test_total_loss = np.mean(total_loss)
    test_ssl_loss = np.mean(total_ssl_loss)
    test_tsk_loss = np.mean(total_tsk_loss)

    return test_total_loss, test_ssl_loss, test_tsk_loss


from sklearn.metrics import roc_auc_score, precision_recall_curve
from sklearn.metrics import auc as prc_auc_score
from sklearn.metrics import f1_score, accuracy_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import matthews_corrcoef


def scorer(y_true, y_pred):

    y_prob = y_pred[:, 1]
    y_pred = y_pred.argmax(axis=1)
    y_true = y_true.argmax(axis=1)
    if len(np.unique(y_true)) == 1:
        roc = np.nan
    else:
        roc = roc_auc_score(y_true, y_prob)
    _precision, _recall, _ = precision_recall_curve(y_true, y_prob)
    prc = prc_auc_score(_recall, _precision)
    f1 = f1_score(y_true, y_pred)
    acc = accuracy_score(y_true, y_pred)
    mcc = matthews_corrcoef(y_true, y_pred)

    return f1, mcc, prc, roc, acc


@torch.no_grad()
def Evaluator(test_loader, model, device):
    model.eval()
    y_trues = []
    y_preds = []
    for data in test_loader:
        triplet, label = data
        anchor, positive, negative = triplet
        anchor_y_true, positive_y_true, negative_y_true = label

        anchor = anchor.to(device)
        anchor_y_true = anchor_y_true.to(device)

        (anchor_emb, anchor_refg), anchor_y_pred = model(anchor)
        anchor_y_pred = torch.nn.functional.softmax(anchor_y_pred, dim=1)

        y_trues.append(anchor_y_true)
        y_preds.append(anchor_y_pred)

    y_true = torch.concat(y_trues, axis=0).cpu().detach().numpy()
    y_pred = torch.concat(y_preds, axis=0).cpu().detach().numpy()

    f1, mcc, prc, roc, acc = scorer(y_true, y_pred)
    return f1, mcc, prc, roc, acc


@torch.no_grad()
def Predictor(dfcx, model, scaler, device="cpu", batch_size=512, num_workers=4):
    model.eval()
    dfcx = scaler.transform(dfcx)

    predict_tcga = GeneData(dfcx)
    predict_loader = Torchdata.DataLoader(
        predict_tcga,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        worker_init_fn=worker_init_fn,
        num_workers=num_workers,
    )
    embds = []
    ys = []
    for anchor in tqdm(predict_loader, ascii=True):
        anchor = anchor.to(device)
        (anchor_emb, anchor_refg), anchor_ys = model(anchor)
        anchor_ys = torch.nn.functional.softmax(anchor_ys, dim=1)
        embds.append(anchor_emb)
        ys.append(anchor_ys)

    embeddings = torch.concat(embds, axis=0).cpu().detach().numpy()
    predictions = torch.concat(ys, axis=0).cpu().detach().numpy()

    if len(model.embed_feature_names) == embeddings.shape[1]:
        columns = model.embed_feature_names
    else:
        columns = model.proj_feature_names

    dfe = pd.DataFrame(embeddings, index=predict_tcga.patient_name, columns=columns)

    ref_cols = model.latentprojector.GENESET.iloc[
        model.latentprojector.CELLPATHWAY.loc["Reference"]
    ].index.tolist()
    ref_cols.append("Reference")

    if not model.ref_for_task:
        dfe = dfe[dfe.columns.difference(ref_cols)]

    dfp = pd.DataFrame(predictions, index=predict_tcga.patient_name)

    return dfe, dfp


@torch.no_grad()
def Extractor(
    dfcx,
    model,
    scaler,
    device="cpu",
    batch_size=512,
    num_workers=4,
    with_gene_level=False,
):
    """
    Extract geneset-level and celltype-level features
    """
    model.eval()
    dfcx = scaler.transform(dfcx)
    genesetprojector = model.latentprojector.genesetprojector
    cellpathwayprojector = model.latentprojector.cellpathwayprojector

    predict_tcga = GeneData(dfcx)
    predict_loader = Torchdata.DataLoader(
        predict_tcga,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        worker_init_fn=worker_init_fn,
        num_workers=num_workers,
    )
    geneset_feat = []
    celltype_feat = []
    gene_feat = []

    for anchor in tqdm(predict_loader, ascii=True):
        anchor = anchor.to(device)
        encoding = model.inputencoder(anchor)

        gene_level_proj = genesetprojector.geneset_scorer(encoding)[
            :, 2:
        ]  # remove pid, cancer
        geneset_level_proj, cellpathway_level_proj = model.latentprojector(encoding)

        geneset_feat.append(geneset_level_proj)
        celltype_feat.append(cellpathway_level_proj)
        gene_feat.append(gene_level_proj)

    genefeatures = torch.concat(gene_feat, axis=0).cpu().detach().numpy()
    genesetfeatures = torch.concat(geneset_feat, axis=0).cpu().detach().numpy()
    celltypefeatures = torch.concat(celltype_feat, axis=0).cpu().detach().numpy()

    dfgeneset = pd.DataFrame(
        genesetfeatures,
        index=predict_tcga.patient_name,
        columns=model.geneset_feature_name,
    )
    dfcelltype = pd.DataFrame(
        celltypefeatures,
        index=predict_tcga.patient_name,
        columns=model.celltype_feature_name,
    )

    dfgene = pd.DataFrame(
        genefeatures, index=predict_tcga.patient_name, columns=predict_tcga.feature_name
    )

    return dfgene, dfgeneset, dfcelltype


@torch.no_grad()
def Projector(dfcx, model, scaler, device="cpu", batch_size=64, num_workers=4):
    """
    Extract geneset-level and celltype-level features
    """

    model.eval()
    gs_projector = model.latentprojector.genesetprojector
    ct_projector = model.latentprojector.cellpathwayprojector

    predict_tcga = GeneData(dfcx)
    predict_loader = Torchdata.DataLoader(
        predict_tcga,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        worker_init_fn=worker_init_fn,
        num_workers=num_workers,
    )
    geneset_feat = []
    celltype_feat = []

    for anchor in tqdm(predict_loader, ascii=True):
        anchor = anchor.to(device)

        x = model.inputencoder(anchor)
        pid_encoding = x[:, 0:1, :]  # take the learnbale patient id token
        cancer_encoding = x[:, 1:2, :]  # take the cancer_type token
        gene_encoding = x[:, 2:, :]  # take the gene encoding

        geneset_level_proj = gs_projector.geneset_aggregator(x)

        b, f, c = geneset_level_proj.shape

        ct_feats = []
        for i in range(c):
            ct_ = ct_projector(geneset_level_proj[:, :, i])
            ct_ = ct_.cpu().detach().numpy()
            ct_feats.append(ct_)

        celltype_level_proj = np.stack(ct_feats, axis=-1)
        geneset_level_proj = geneset_level_proj.cpu().detach().numpy()

        geneset_feat.append(geneset_level_proj)
        celltype_feat.append(celltype_level_proj)

    gs_feat = np.concatenate(geneset_feat, axis=0)
    ct_feat = np.concatenate(celltype_feat, axis=0)

    b, f, c = gs_feat.shape
    feature_name = model.latentprojector.GENESET.index
    feature_sample_labels = [
        dfcx.index[i // f] + "$$" + feature_name[i % f] for i in range(b * f)
    ]
    dfgs = pd.DataFrame(
        gs_feat.reshape(b * f, c),
        index=feature_sample_labels,
        columns=["channel_%s" % i for i in range(c)],
    )

    b, f, c = ct_feat.shape
    feature_name = model.latentprojector.CELLPATHWAY.index
    index = [dfcx.index[i // f] + "$$" + feature_name[i % f] for i in range(b * f)]
    columns = ["channel_%s" % i for i in range(c)]
    dfct = pd.DataFrame(ct_feat.reshape(b * f, c), index=index, columns=columns)
    return dfgs, dfct



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 15:05:13 2023

@author: Wanxiang Shen
"""

import os
import numpy as np
import pandas as pd
from glob import glob
import torch
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import Dataset
from torch.nn.functional import normalize
from torch.distributions import Beta


def tolist(x):
    return x.tolist()


class NoScaler:
    """A scaler that performs no scaling (identity)."""

    def __init__(self):
        pass

    def fit(self, dfx):
        return self

    def transform(self, dfx):
        return dfx

    def fit_transform(self, dfx):
        return dfx

    def inverse_transform(self, dfx):
        return dfx


class TCGAData(Dataset):
    def __init__(self, dfcx, dfy, augmentor, K=1):
        """
        Parameters:
            dfcx (DataFrame): Cancer type and TPM matrix, first column is cancer type (0-32).
            dfy (DataFrame): Task dataframe, one-hot encoding for classification.
            augmentor: Instance of MixupNormalAugmentor.
            K (float): Determines the ratio of nearest neighbors used.
        """
        self.feature_name = dfcx.columns[1:]  # first column is cancer type
        self.patient_name = dfcx.index
        self.task_cols = dfy.columns
        self.task_dim = len(dfy.columns)

        self.cancer_type = dfcx[dfcx.columns[0]].values
        
        X = torch.tensor(dfcx.values, dtype=torch.float32).clone().detach()
        self.X = X

        y_scaler = MinMaxScaler()
        y = y_scaler.fit_transform(dfy.values)
        self.y = torch.tensor(y, dtype=torch.float32).clone().detach()
        self.y_scaler = y_scaler

        self.K = K

        dist = torch.cdist(X[:, 1:], X[:, 1:])
        if K in [-1, 1, 0]:
            knn_value, knn_idx = dist.topk(len(X), largest=False)
        else:
            knn_value, knn_idx = dist.topk(int(len(X) * K), largest=False)

        self.knn_value = knn_value[:, 1:]
        self.knn_idx = knn_idx[:, 1:]

        self.augmentor = augmentor

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        # anchor sample
        a = self.X[idx]
        topK_idx = self.knn_idx[idx]

        topK_idx = [i for i in topK_idx if self.cancer_type[i] != self.cancer_type[idx]]  # filter the same cancer type
        
        # positive sample augmentation
        xa = self.augmentor.augment_a(a)[0]
        xp = self.augmentor.augment_p(a)[0]

        # negative sample selection
        neg_idx = np.random.choice(topK_idx)
        n = self.X[neg_idx]

        ya = self.y[idx]
        yp = ya.clone()
        yn = self.y[neg_idx]

        xn = self.augmentor.augment_n(n)[0]

        x = [xa, xp, xn]
        y = [ya, yp, yn]

        return x, y


class GeneData(Dataset):
    def __init__(self, dfcx):
        """
        Parameters:
            dfcx (DataFrame): TPM matrix.
        """
        self.feature_name = dfcx.columns[1:]
        self.patient_name = dfcx.index

        X = torch.tensor(dfcx.values, dtype=torch.float32).clone().detach()
        self.X = X
        self.dfcx = dfcx

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        # single sample retrieval
        x = self.X[idx]
        return x


class ITRPData(Dataset):
    def __init__(self, dfcx, dfy):
        """
        Parameters:
            dfcx (DataFrame): TPM matrix.
            dfy (DataFrame): Task dataframe with response labels (one-hot encoded).
        """
        self.feature_name = dfcx.columns[1:]
        self.patient_name = dfcx.index
        self.task_cols = dfy.columns  # ['NR', 'R']
        self.task_dim = len(dfy.columns)

        X = torch.tensor(dfcx.values, dtype=torch.float32).clone().detach()
        self.X = X

        y_scaler = MinMaxScaler()
        y = y_scaler.fit_transform(dfy.values)
        self.y = torch.tensor(y, dtype=torch.float32).clone().detach()
        self.y_scaler = y_scaler

        self.responder_idx = torch.where(self.y[:, 1] == 1.0)[0]
        self.nonresponder_idx = torch.where(self.y[:, 1] == 0.0)[0]

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        ya = self.y[idx]

        # select positive and negative samples based on response
        if ya[1] == 1:
            pos_idx = np.random.choice(self.responder_idx)
            neg_idx = np.random.choice(self.nonresponder_idx)
        else:
            pos_idx = np.random.choice(self.nonresponder_idx)
            neg_idx = np.random.choice(self.responder_idx)

        xa = self.X[idx]
        xp = self.X[pos_idx]
        xn = self.X[neg_idx]

        x = [xa, xp, xn]

        yp = self.y[pos_idx]
        yn = self.y[neg_idx]

        y = [ya, yp, yn]

        return x, y



from .cellpathwayaggregator import CellPathwayAggregator
from .genesetaggregator import GeneSetAggregator
from .genesetscorer import GeneSetScorer
from ..tokenizer import CONCEPT, TOKENS_SHORT, TOKENS_LONG


import torch, math
import torch.nn as nn
import torch.nn.functional as F
import os
import pandas as pd
import numpy as np


cwd = os.path.dirname(__file__)


class GeneSetProjector(nn.Module):
    def __init__(
        self,
        GENESET,
        geneset_feature_dim,
        geneset_agg_mode="attention",
        geneset_score_mode="linear",
    ):
        super(GeneSetProjector, self).__init__()

        self.geneset_feature_dim = geneset_feature_dim
        self.geneset_agg_mode = geneset_agg_mode
        self.geneset_score_mode = geneset_score_mode

        self.GENESET = GENESET
        self.genesets_indices = GENESET.tolist()
        self.genesets_names = GENESET.index

        self.geneset_aggregator = GeneSetAggregator(
            self.genesets_indices, self.geneset_agg_mode
        )
        self.geneset_scorer = GeneSetScorer(
            self.geneset_feature_dim, self.geneset_score_mode
        )

        # self.norm = nn.LayerNorm(len(self.genesets_indices))
        # nn.BatchNorm1d(n)

    def forward(self, x):
        geneset_feats = self.geneset_aggregator(x)

        geneset_scores = self.geneset_scorer(geneset_feats)
        # geneset_scores = F.normalize(geneset_scores, p=2., dim=1)
        return geneset_scores


class CellPathwayProjector(nn.Module):

    def __init__(self, CELLPATHWAY, cellpathway_agg_mode="pooling"):
        super(CellPathwayProjector, self).__init__()
        self.cellpathway_agg_mode = cellpathway_agg_mode
        self.CELLPATHWAY = CELLPATHWAY
        self.cellpathway_indices = CELLPATHWAY.tolist()
        self.cellpathway_names = CELLPATHWAY.index
        self.cellpathway_aggregator = CellPathwayAggregator(
            self.cellpathway_indices, mode=self.cellpathway_agg_mode
        )

    def forward(self, x):
        cellpathway_scores = self.cellpathway_aggregator(x)
        return cellpathway_scores


class DisentangledProjector(nn.Module):

    def __init__(
        self,
        gene_num,
        gene_feature_dim,
        proj_pid=True,
        proj_cancer_type=True,
        geneset_agg_mode="attention",
        geneset_score_mode="linear",
        cellpathway_agg_mode="attention",
    ):

        super(DisentangledProjector, self).__init__()

        TOKENS = TOKENS_LONG if len(TOKENS_LONG) == gene_num + 2 else TOKENS_SHORT
        TOKEN_IDX_MAP = (
            pd.Series(TOKENS).reset_index().set_index(0)["index"].astype(int).to_dict()
        )

        def _genes_to_idxs(genes):
            TOKEN_IDX_MAP
            gene_idxs = []
            for i in genes.split(":"):
                idx = TOKEN_IDX_MAP.get(i)
                if idx != None:
                    gene_idxs.append(idx)
            return gene_idxs

        GENESET = CONCEPT.Genes.apply(_genes_to_idxs)
        CELLPATHWAY = (
            CONCEPT.reset_index()
            .groupby("BroadCelltypePathway")
            .apply(lambda x: x.index.tolist())
        )

        ##sort the concept by the order
        GENESET = GENESET.loc[CONCEPT.sort_values("GeneSet_index").index]
        c_order = (
            CONCEPT[["BroadCelltypePathway", "Concept_index"]]
            .drop_duplicates()
            .sort_values("Concept_index")["BroadCelltypePathway"]
            .tolist()
        )
        CELLPATHWAY = CELLPATHWAY.loc[c_order]

        assert "Reference" in CELLPATHWAY.index, "Reference must in CELLPATHWAY!"
        # Move 'Reference' to the end
        noref = CELLPATHWAY[CELLPATHWAY.index != "Reference"]
        ref = CELLPATHWAY[CELLPATHWAY.index == "Reference"]
        CELLPATHWAY = noref._append(ref)

        self.GENESET = GENESET
        self.CELLPATHWAY = CELLPATHWAY

        geneset_name_list = GENESET.index.tolist()
        celltype_name_list = CELLPATHWAY.index.tolist()

        self.proj_pid = proj_pid
        self.proj_cancer_type = proj_cancer_type

        self.gene_num = gene_num
        self.gene_feature_dim = gene_feature_dim
        self.geneset_agg_mode = geneset_agg_mode
        self.geneset_score_mode = geneset_score_mode
        self.cellpathway_agg_mode = cellpathway_agg_mode
        self.genesetprojector = GeneSetProjector(
            GENESET,
            self.gene_feature_dim,
            self.geneset_agg_mode,
            self.geneset_score_mode,
        )
        self.cellpathwayprojector = CellPathwayProjector(
            CELLPATHWAY, self.cellpathway_agg_mode
        )

        self.patientprojector = GeneSetScorer(self.gene_feature_dim, "linear")
        self.cancerprojector = GeneSetScorer(self.gene_feature_dim, "linear")

        if proj_pid and proj_cancer_type:
            PROJCOLS = ["PID", "CANCER"]

        elif proj_pid and not proj_cancer_type:
            PROJCOLS = ["PID"]

        elif not proj_pid and proj_cancer_type:
            PROJCOLS = ["CANCER"]

        else:
            PROJCOLS = []

        geneset_proj_cols = PROJCOLS.copy()
        geneset_proj_cols.extend(geneset_name_list)

        cellpathway_proj_cols = PROJCOLS.copy()
        cellpathway_proj_cols.extend(celltype_name_list)

        self.cellpathway_proj_cols = cellpathway_proj_cols
        self.geneset_proj_cols = geneset_proj_cols

        ## reference genes
        ref_gene_ids = []
        for refset in GENESET.iloc[CELLPATHWAY["Reference"]].to_list():
            ref_gene_ids.extend(refset)

        ## reference sets
        ref_geneset_ids = []
        for rgs in GENESET.iloc[CELLPATHWAY["Reference"]].index:
            idx = geneset_proj_cols.index(rgs)
            ref_geneset_ids.append(idx)

        ## reference cellpathways
        ref_celltype_ids = [cellpathway_proj_cols.index("Reference")]

        self.ref_gene_ids = ref_gene_ids
        self.ref_celltype_ids = ref_celltype_ids
        self.ref_geneset_ids = ref_geneset_ids

    def forward(self, x):

        # x size is  B, L+2, 1
        pid_encoding = x[:, 0:1, :]  # take the learnbale patient id token
        cancer_encoding = x[:, 1:2, :]  # take the cancer_type token
        gene_encoding = x[:, 2:, :]  # take the gene encoding

        geneset_scores = self.genesetprojector(gene_encoding)
        cellpathway_scores = self.cellpathwayprojector(geneset_scores)

        if self.proj_pid and self.proj_cancer_type:
            cancer_scores = self.cancerprojector(cancer_encoding)
            pid_scores = self.patientprojector(pid_encoding)
            geneset_proj = torch.cat([pid_scores, cancer_scores, geneset_scores], dim=1)
            cellpathway_proj = torch.cat(
                [pid_scores, cancer_scores, cellpathway_scores], dim=1
            )

        elif self.proj_pid and not self.proj_cancer_type:
            pid_scores = self.patientprojector(pid_encoding)
            geneset_proj = torch.cat([pid_scores, geneset_scores], dim=1)
            cellpathway_proj = torch.cat([pid_scores, cellpathway_scores], dim=1)

        elif not self.proj_pid and self.proj_cancer_type:
            cancer_scores = self.cancerprojector(cancer_encoding)
            geneset_proj = torch.cat([cancer_scores, geneset_scores], dim=1)
            cellpathway_proj = torch.cat([cancer_scores, cellpathway_scores], dim=1)

        else:
            geneset_proj = geneset_scores
            cellpathway_proj = cellpathway_scores

        return geneset_proj, cellpathway_proj


class GeneSetPlaceholderAggregator(nn.Module):
    def __init__(self, genes_num, genesets_num):
        """
        Initializes the GeneSetPlaceholderAggregator module.
        :param gene_sets: A list of lists, where each inner list contains indices of genes in a gene set.

        """
        super(GeneSetPlaceholderAggregator, self).__init__()

        self.genesets_num = genesets_num
        self.genes_num = genes_num

        # Attention weights for each gene in each placeholder gene set
        self.attention_weights = nn.ParameterDict(
            {
                f"Placeholder_geneset_{i}": nn.Parameter(torch.randn(genes_num, 1))
                for i, gene_set in enumerate(range(genesets_num))
            }
        )

    def forward(self, x):
        transformed = self.transformer(x)
        gene_set_outputs = []
        for i in range(self.genesets_num):
            weight = self.attention_weights[f"Placeholder_geneset_{i}"]
            gene_set_output = transformed * weight
            gene_set_outputs.append(gene_set_output.sum(dim=1))

        return torch.stack(gene_set_outputs, dim=1)

    def adaptive_prune(self):
        with torch.no_grad():
            all_weights = torch.cat(
                [param.view(-1) for param in self.attention_weights.values()]
            )
            threshold = torch.quantile(torch.abs(all_weights), 0.90)  # 保留最重要的10%
            for name, param in self.attention_weights.items():
                mask = torch.abs(param) >= threshold
                param.mul_(mask)


## need to be revised to fit the cls and cancer type token
class EntangledProjector(nn.Module):
    def __init__(self, gene_feature_dim, mode="mean"):
        """
        reduce: {'mean', 'max', 'cls'}
        """
        super(EntangledProjector, self).__init__()
        self.mode = mode
        self.gene_feature_dim = gene_feature_dim

    def forward(self, x):
        if self.mode == "mean":
            x = torch.mean(x, dim=-1)
        elif self.mode == "max":
            x = torch.max(x, dim=-1)[0]
        else:
            ValueError("Invalid pooling type. Use 'mean' or 'max'.")
        return x



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 16:44:09 2023

@author: Wanxiang Shen
"""

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(style="white", font_scale=1.5)
from tqdm import tqdm
import torch
import wandb

import torch.utils.data as data
from copy import deepcopy
from collections import OrderedDict
from joblib import dump, load
from sklearn.model_selection import StratifiedKFold

import random
import datetime


from compass.dataloader import TCGAData, GeneData, ITRPData
from compass.augmentor import (
    RandomMaskAugmentor,
    FeatureJitterAugmentor,
    MaskJitterAugmentor,
)
from compass.model.scaler import Datascaler
from compass.model.model import Compass
from compass.model.train import PT_Trainer, PT_Tester
from compass.model.adapt import Adp_Trainer, Adp_Tester

from compass.model.tune import (
    FT_Trainer,
    FT_Tester,
    Predictor,
    Evaluator,
    Extractor,
    Projector,
)

from compass.model.loss import TripletLoss, CEWithNaNLabelsLoss, MAEWithNaNLabelsLoss
from compass.model.loss import FocalLoss, DiceLoss, DSCLoss, HingeLoss
from compass.model.saver import SaveBestModel
from compass.utils import (
    plot_embed_with_label,
    loadcompass,
)  #  code for compass model loading

def fixseed(seed=42):
    np.random.seed(seed)
    random.seed(seed)
    if seed is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


# fixseed(seed=42)
def freeze_bn_dropout(m):
    if isinstance(m, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d, 
                      torch.nn.BatchNorm3d, torch.nn.Dropout)):
        m.eval()

def worker_init_fn(worker_id):
    seed = torch.initial_seed() % 2**32
    np.random.seed(seed)


class PreTrainer:

    def __init__(
        self,
        device="cuda",
        lr=1e-4,
        weight_decay=1e-10,
        epochs=100,
        patience=10,
        batch_size=128,
        embed_dim=32,
        task_loss_weight=0.0,
        task_dense_layer=[24],
        task_batch_norms=True,
        task_class_weight=None,
        encoder="performer",
        encoder_dropout=0.2,
        num_cancer_types=33,
        transformer_dim=32,
        transformer_num_layers=1,
        transformer_nhead=2,
        transformer_pos_emb="learnable",
        batch_correction=0.0,
        proj_level="cellpathway",
        proj_pid=False,
        proj_cancer_type=True,
        ref_for_task=True,
        triplet_metric="cosine",
        triplet_margin=1.0,
        K=1,
        seed=64,
        work_dir="./results",
        verbose=True,
        with_wandb=False,
        wandb_project="pretrain",
        wandb_dir="./wandb/",
        wandb_entity="senwanxiang",
        **encoder_kwargs,
    ):
        """
        transformer_pos_emb: {None, 'umap', 'pumap', 'gene2vect'}
        encoder:{'mlp', 'transformer'}
        """

        self.device = device
        self.lr = lr
        self.weight_decay = weight_decay
        self.epochs = epochs
        self.batch_size = batch_size
        self.embed_dim = embed_dim
        self.num_cancer_types = num_cancer_types

        self.triplet_margin = triplet_margin
        self.triplet_metric = triplet_metric

        self.K = K

        self.task_loss_weight = task_loss_weight
        self.task_dense_layer = task_dense_layer
        self.task_batch_norms = task_batch_norms
        self.task_class_weight = task_class_weight
        self.encoder = encoder
        self.encoder_dropout = encoder_dropout
        self.transformer_dim = transformer_dim
        self.transformer_nhead = transformer_nhead
        self.transformer_num_layers = transformer_num_layers
        self.transformer_pos_emb = transformer_pos_emb

        self.batch_correction = batch_correction
        self.proj_level = proj_level
        self.proj_pid = proj_pid
        self.proj_cancer_type = proj_cancer_type
        self.ref_for_task = ref_for_task

        fixseed(seed=seed)
        self.seed = seed
        self.patience = patience  # for early stopping

        self.work_dir = work_dir
        self.with_wandb = with_wandb
        self.verbose = verbose
        self.wandb_project = wandb_project
        self.wandb_dir = wandb_dir
        self.wandb_entity = wandb_entity
        self.encoder_kwargs = encoder_kwargs

    def _setup(self, input_dim, task_dim, task_type, save_dir, run_name):

        model = Compass(
            input_dim,
            task_dim,
            task_type,
            proj_level=self.proj_level,
            proj_pid=self.proj_pid,
            proj_cancer_type=self.proj_cancer_type,
            ref_for_task=self.ref_for_task,
            num_cancer_types=self.num_cancer_types,
            embed_dim=self.embed_dim,
            encoder=self.encoder,
            encoder_dropout=self.encoder_dropout,
            transformer_dim=self.transformer_dim,
            transformer_nhead=self.transformer_nhead,
            transformer_num_layers=self.transformer_num_layers,
            transformer_pos_emb=self.transformer_pos_emb,
            task_dense_layer=self.task_dense_layer,
            task_batch_norms=self.task_batch_norms,
            seed=self.seed,
            **self.encoder_kwargs,
        )

        model = model.to(self.device)
        self.model = model
        
        ssl_loss = TripletLoss(margin=self.triplet_margin, metric=self.triplet_metric)
        ce_loss = CEWithNaNLabelsLoss(weights=self.task_class_weight)
        mae_loss = MAEWithNaNLabelsLoss()

        optimizer = torch.optim.Adam(model.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        saver = SaveBestModel(save_dir=save_dir, save_name="model.pth")

        if task_type == "c":
            tsk_loss = ce_loss
        elif task_type == "r":
            tsk_loss = mae_loss
        else:
            raise ValueError("Invalid task_type. Use 'c', or 'r'. ")

        self.ssl_loss = ssl_loss
        self.tsk_loss = tsk_loss
        self.optimizer = optimizer
        self.saver = saver

        if self.with_wandb:
            # Initialize wandb
            self.wandb = wandb.init(
                project=self.wandb_project,
                entity=self.wandb_entity,
                dir=self.wandb_dir,
                name=run_name,
                save_code=True,
            )

            # # Log model architecture to wandb
            self.wandb.watch(model)

    def train(
        self,
        dfcx_train,
        dfy_train=None,
        dfcx_test=None,
        dfy_test=None,
        task_name="notask",
        task_type="c",
        aug_method="mix",
        scale_method="minmax",
        **augargs,
    ):

        ### scaler ####
        self.scale_method = scale_method
        self.scaler = Datascaler(scale_method=scale_method)
        self.scaler = self.scaler.fit(dfcx_train)
        self.aug_method = aug_method

        dfcx_train = self.scaler.transform(dfcx_train)

        if aug_method == "mask":
            self.augmentor = RandomMaskAugmentor(**augargs)
        elif aug_method == "jitter":
            self.augmentor = FeatureJitterAugmentor(**augargs)
        elif aug_method == "mix":
            self.augmentor = MaskJitterAugmentor(**augargs)
        else:
            raise ValueError("Invalid method. Use 'mask', 'jitter' or 'mix'.")

        self.task_type = task_type
        self.task_name = task_name

        if dfy_train is None:
            dfy_train = pd.DataFrame(
                np.zeros((len(dfcx_train), 2)), index=dfcx_train.index
            )

        train_tcga = TCGAData(dfcx_train, dfy_train, self.augmentor, K=self.K)

        self.y_scaler = train_tcga.y_scaler
        self.feature_name = train_tcga.feature_name

        train_loader = data.DataLoader(
            train_tcga,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
            pin_memory=True,
            num_workers=4,
        )

        input_dim = len(train_tcga.feature_name)
        task_dim = train_tcga.y.shape[1]

        if dfcx_test is not None:
            dfcx_test = self.scaler.transform(dfcx_test)
            if dfy_test is None:
                dfy_test = pd.DataFrame(
                    np.zeros((len(dfcx_test), 2)), index=dfcx_test.index
                )
            test_tcga = TCGAData(dfcx_test, dfy_test, self.augmentor, K=self.K)
            test_loader = data.DataLoader(
                test_tcga,
                batch_size=self.batch_size,
                shuffle=False,
                pin_memory=True,
                num_workers=4,
            )
        else:
            test_loader = None

        current_time = datetime.datetime.now()
        formatted_time = current_time.strftime(
            "%Y%m%d_%H%M%S"
        )  # Format: YYYYMMDD_HHMMSS
        run_name = f"Pretrain_{self.task_name}_{formatted_time}"  # Example filename

        self.run_name = run_name
        self.save_dir = os.path.join(self.work_dir, run_name)

        ## init model, operimizer, ...
        self._setup(input_dim, task_dim, task_type, self.save_dir, self.run_name)
        self.input_dim = input_dim
        self.task_dim = task_dim

        ### training ###
        performace = []
        best_val_loss = float("inf")
        patience_counter = 0
        for epoch in range(self.epochs):
            train_total_loss, train_ssl_loss, train_tsk_loss = PT_Trainer(
                train_loader,
                self.model,
                self.optimizer,
                self.ssl_loss,
                self.tsk_loss,
                self.device,
                alpha=self.task_loss_weight,
                correction=self.batch_correction,
            )

            if test_loader is not None:
                test_total_loss, test_ssl_loss, test_tsk_loss = PT_Tester(
                    test_loader,
                    self.model,
                    self.ssl_loss,
                    self.tsk_loss,
                    self.device,
                    alpha=self.task_loss_weight,
                    correction=self.batch_correction,
                )

                self.saver(
                    test_total_loss, epoch, self.model, self.optimizer, self.scaler
                )

                # Early Stopping Check
                if test_total_loss < best_val_loss:
                    best_val_loss = test_total_loss
                    patience_counter = 0  # Reset the counter if improvement is seen
                else:
                    patience_counter += 1  # Increment the counter if no improvement

            else:
                test_total_loss, test_ssl_loss, test_tsk_loss = np.nan, np.nan, np.nan
                self.saver(
                    train_total_loss, epoch, self.model, self.optimizer, self.scaler
                )

                # Early Stopping Check
                if train_total_loss < best_val_loss:
                    best_val_loss = train_total_loss
                    patience_counter = 0  # Reset the counter if improvement is seen
                else:
                    patience_counter += 1  # Increment the counter if no improvement

            performace.append(
                [
                    epoch,
                    train_total_loss,
                    train_ssl_loss,
                    train_tsk_loss,
                    test_total_loss,
                    test_ssl_loss,
                    test_tsk_loss,
                ]
            )

            if self.with_wandb:
                self.wandb.log(
                    {
                        "pretrain_loss": train_total_loss,
                        "pretrain_ssl_loss": train_ssl_loss,
                        "pretrain_%s_loss" % self.task_name: train_tsk_loss,
                        "pretrain_test_loss": test_total_loss,
                        "pretrain_test_ssl_loss": test_ssl_loss,
                        "pretrain_test_%s_loss" % self.task_name: test_tsk_loss,
                    }
                )
            if self.verbose:
                print(
                    "Epoch: {}/{} - Train Loss: {:.4f} - Test Loss: {:.4f}".format(
                        epoch + 1, self.epochs, train_total_loss, test_total_loss
                    )
                )

            # for early stopping
            if patience_counter >= self.patience:
                print(
                    f"Stopping early at epoch {epoch+1}. No improvement in validation loss for {self.patience} consecutive epochs."
                )
                break

        self.saver.save()
        self.performace = performace

        ## plot loss locally
        df = pd.DataFrame(
            self.performace,
            columns=[
                "epochs",
                "total_loss",
                "ssl_loss",
                "%s_loss" % self.task_name,
                "test_loss",
                "test_ssl_loss",
                "test_%s_loss" % self.task_name,
            ],
        ).set_index("epochs")
        # v = (df - df.min(axis=0)) / (df.max(axis=0) - df.min(axis=0))
        fig, ax = plt.subplots(figsize=(7, 5))
        df[["ssl_loss", "test_ssl_loss"]].plot(ax=ax)
        fig.savefig(
            os.path.join(self.save_dir, "tcga_train_loss.png"), bbox_inches="tight"
        )
        df.to_pickle(os.path.join(self.save_dir, "tcga_train_loss.pkl"))

        return self

    def plot_embed(self, df_tpm, df_label, label_types, **kwargs):
        dfe, dfp = self.predict(df_tpm, batch_size=self.batch_size, num_workers=4)
        dfd = dfe.join(df_label)
        label_cols = df_label.columns

        figs = plot_embed_with_label(
            dfd, label_col=label_cols, label_type=label_types, **kwargs
        )
        for fig, name in zip(figs, label_cols):
            fig.savefig(
                os.path.join(self.save_dir, "tcga_%s.png" % name),
                bbox_inches="tight",
            )

    def predict(self, df_tpm, batch_size=512, num_workers=4):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfe, dfp = Predictor(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        return dfe, dfp

    def extract(self, df_tpm, batch_size=512, num_workers=4, with_gene_level=False):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfg, dfgs, dfct = Extractor(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
            with_gene_level=True,
        )
        if with_gene_level:
            return dfg, dfgs, dfct
        else:
            return dfgs, dfct

    def project(self, df_tpm, batch_size=512, num_workers=4):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfg, dfc = Projector(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        return dfg, dfc

    def save(self, mfile):
        if self.with_wandb:
            self.wandb._settings = ""
        torch.save(self, mfile)
        print("Saving the model to %s" % mfile)

    def load(self, mfile, **kwargs):
        self = loadcompass(mfile, **kwargs)
        if self.with_wandb:
            self.wandb._settings = ""
        return self

    def count_parameters(self):
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)

    def close(self):
        if self.with_wandb:
            self.wandb.finish()
            self.wandb._settings = ""
        self.save(os.path.join(self.save_dir, "pretrainer.pt"))

    def copy(self):
        return deepcopy(self)


class Adapter:

    def __init__(
        self,
        pretrainer,
        adp_feature="TMB",
        lr=1e-3,
        weight_decay=1e-6,
        batch_size=128,
        epochs=100,
        patience=10,
        save_dir="./",
        verbose=True,
        with_wandb=False,
        save_best_model=False,
    ):
        self.lr = lr
        self.weight_decay = weight_decay
        self.epochs = epochs
        self.patience = patience
        self.pretrainer = pretrainer.copy()
        self.adp_feature = adp_feature
        self.batch_size = batch_size
        self.save_dir = save_dir
        self.with_wandb = with_wandb

        self._init_adaptive_model(Key=self.adp_feature)
        self.verbose = verbose
        self.save_best_model = save_best_model
        self.scaler = pretrainer.scaler
        self.device = pretrainer.device

    def _init_adaptive_model(self, Key="TMB"):

        fixseed(self.pretrainer.seed)  # Resets the random seed
        model_args = deepcopy(self.pretrainer.saver.inMemorySave["model_args"])
        model_weights = deepcopy(self.pretrainer.saver.inMemorySave["model_state_dict"])

        ### define finetune model
        model = Compass(**model_args)

        encoder_state = OrderedDict()
        for k, v in model_weights.items():
            encoder_state[k] = v

        ### load Pretrained model
        model.load_state_dict(encoder_state, strict=False)
        model = model.to(self.pretrainer.device)

        ctp_idx = (
            model.latentprojector.cellpathwayprojector.CELLPATHWAY.index.tolist().index(
                Key
            )
        )
        proj_idx = model.latentprojector.cellpathway_proj_cols.index(Key)

        ctp_name = "cellpathway_%s" % ctp_idx

        gs_idx_list = model.latentprojector.cellpathwayprojector.CELLPATHWAY.loc[Key]
        gs_name_list = ["geneset_%s" % idx for idx in gs_idx_list]

        for key, param in model.latentprojector.cellpathwayprojector.named_parameters():
            key_idx = key.split(".")[-1]
            if key_idx == ctp_name:
                print(key)
                param.requires_grad = True
            else:
                param.requires_grad = False

        for key, param in model.latentprojector.genesetprojector.named_parameters():
            key_idx = key.split(".")[-1]
            if key_idx in gs_name_list:
                print(key)
                param.requires_grad = True
            else:
                param.requires_grad = False

        for param in model.inputencoder.parameters():
            param.requires_grad = False

        for param in model.latentprojector.patientprojector.parameters():
            param.requires_grad = False

        for param in model.latentprojector.cancerprojector.parameters():
            param.requires_grad = False

        for param in model.taskdecoder.parameters():
            param.requires_grad = False

        plist = []
        for param in model.parameters():
            if param.requires_grad:
                plist.append({"params": param})

        optimizer = torch.optim.Adam(plist, lr=self.lr, weight_decay=self.weight_decay)
        loss = MAEWithNaNLabelsLoss()
        saver = SaveBestModel(save_dir=self.save_dir, save_name="adp_model.pth")

        self.model = model
        self.proj_idx = proj_idx
        self.optimizer = optimizer
        self.loss = loss
        self.saver = saver
        plist.append({"params": param})

    def adapt(self, dfcx_train, dfy_train, dfcx_test=None, dfy_test=None):

        dfcx_train = self.scaler.transform(dfcx_train)

        augmentor = RandomMaskAugmentor(no_augment_prob=1)  # no aug

        train_itrp = TCGAData(dfcx_train, dfy_train, augmentor, K=1)
        train_loader = data.DataLoader(
            train_itrp,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=False,
            worker_init_fn=worker_init_fn,
            pin_memory=True,
            num_workers=4,
        )  #

        if dfcx_test is not None:
            dfcx_test = self.scaler.transform(dfcx_test)
            test_itrp = TCGAData(dfcx_test, dfy_test, augmentor, K=1)
            test_loader = data.DataLoader(
                test_itrp,
                batch_size=self.batch_size,
                shuffle=False,
                worker_init_fn=worker_init_fn,
                pin_memory=True,
                num_workers=4,
            )
        else:
            test_loader = None

        ### training ###
        performance = []
        best_val_loss = float("inf")
        patience_counter = 0
        for epoch in tqdm(range(self.epochs), ascii=True):

            train_total_loss = Adp_Trainer(
                train_loader,
                self.model,
                self.optimizer,
                self.loss,
                self.device,
                self.proj_idx,
            )

            saving_criteria = train_total_loss

            if test_loader is not None:
                test_total_loss = Adp_Tester(
                    test_loader,
                    self.model,
                    self.optimizer,
                    self.loss,
                    self.device,
                    self.proj_idx,
                )

                saving_criteria = test_total_loss

            else:
                test_total_loss = np.nan

            ## saving_criteria: lower is better
            ## saving model & Early Stopping based on saving_criteria
            self.saver(saving_criteria, epoch, self.model, self.optimizer, self.scaler)
            if saving_criteria < best_val_loss:
                best_val_loss = saving_criteria
                patience_counter = 0
            else:
                patience_counter += 1

            performance.append([epoch, train_total_loss, test_total_loss])

            if self.with_wandb:
                self.wandb.log(
                    {"train_loss": train_total_loss, "test_loss": test_total_loss}
                )

            if self.verbose:
                print(
                    "Epoch: {}/{} - Train Loss: {:.4f} - Test Loss: {:.4f}".format(
                        epoch + 1, self.epochs, train_total_loss, test_total_loss
                    )
                )

            if patience_counter >= self.patience:
                print(
                    f"Stopping early at epoch {epoch+1}. No improvement in validation loss for {self.patience} consecutive epochs."
                )
                break

        if self.save_best_model:
            self.saver.save()
        else:
            os.system('rm -r "%s"' % self.save_dir)

        self.performance = performance
        self.best_epoch = self.saver.inMemorySave["epoch"]

        self.pretrainer.saver = self.saver
        self.pretrainer.model = self.model

        return self.pretrainer

    def count_parameters(self):
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)


class FineTuner:
    """
    Few-shot learning (with <full, partial, head> or without FT) on ITRP datasets
    """

    def __init__(
        self,
        pretrainer,
        mode="PFT",
        load_decoder=False,
        device="cuda",
        lr=1e-3,
        weight_decay=1e-4,
        max_epochs=500,
        patience=10,
        batch_size=32,
        triplet_metric="cosine",
        triplet_margin=1.0,
        task_loss_weight=1.0,
        task_loss_type="ce_loss",
        task_class_weight=[1, 2],
        task_dense_layer=[16],
        task_batch_norms=True,
        batch_correction=0.0,
        entropy_weight=0.0,
        seed=42,
        verbose=True,
        with_wandb=False,
        wandb_project="finetune",
        wandb_dir="/n/data1/hms/dbmi/zitnik/lab/users/was966/wandb/",
        wandb_entity="senwanxiang",
        work_dir="./results",
        task_name="rps",
        task_type="c",
        task_dim=2,
        save_best_model=False,
    ):
        """
        pretrainer: TCGAPreTrainer
        mode: tuning mode{head:LFT, partial: PFT, or full: FFT}
        """

        self.pretrainer = pretrainer.copy()
        self.scaler = self.pretrainer.scaler

        self.mode = mode
        self.load_decoder = load_decoder

        self.device = device
        self.lr = lr
        self.weight_decay = weight_decay
        self.max_epochs = max_epochs
        self.patience = patience
        self.batch_size = batch_size
        self.triplet_metric = triplet_metric
        self.triplet_margin = triplet_margin

        self.task_loss_weight = task_loss_weight
        self.task_loss_type = task_loss_type

        self.task_dense_layer = task_dense_layer
        self.task_batch_norms = task_batch_norms
        self.task_class_weight = task_class_weight
        self.batch_correction = batch_correction

        self.entropy_weight = entropy_weight
        self.seed = seed
        fixseed(self.seed)

        self.work_dir = work_dir
        self.with_wandb = with_wandb
        self.wandb_project = wandb_project
        self.wandb_dir = wandb_dir
        self.wandb_entity = wandb_entity
        self.verbose = verbose

        self.task_type = task_type
        self.task_name = task_name
        self.task_dim = task_dim
        self.save_best_model = save_best_model

        self.params = {
            "mode": self.mode,
            "load_decoder": self.load_decoder,
            "device": self.device,
            "lr": self.lr,
            "weight_decay": self.weight_decay,
            "max_epochs": self.max_epochs,
            "patience": self.patience,
            "batch_size": self.batch_size,
            "triplet_margin": self.triplet_margin,
            "triplet_metric": self.triplet_metric,
            "entropy_weight": self.entropy_weight,
            "task_loss_weight": self.task_loss_weight,
            "task_loss_type": self.task_loss_type,
            "task_dense_layer": self.task_dense_layer,
            "task_batch_norms": self.task_batch_norms,
            "task_class_weight": self.task_class_weight,
            "batch_correction": self.batch_correction,
            "work_dir": self.work_dir,
            "with_wandb": self.with_wandb,
            "wandb_project": self.wandb_project,
            "wandb_dir": self.wandb_dir,
            "wandb_entity": self.wandb_entity,
            "verbose": self.verbose,
            "task_dim": self.task_dim,
            "task_type": self.task_type,
            "save_best_model": self.save_best_model,
        }

        __all_loss_type = [
            "ce_loss",
            "focal_loss",
            "mae_loss",
            "dice_loss",
            "dsc_loss",
            "hinge_loss",
        ]
        assert task_loss_type in __all_loss_type, (
            "Unsupported loss type, please input one of %" % __all_loss_type
        )

        current_time = datetime.datetime.now()
        formatted_time = current_time.strftime(
            "%Y%m%d_%H%M%S"
        )  # Format: YYYYMMDD_HHMMSS
        run_name = f"Finetune_{self.task_name}_{formatted_time}"  # Example filename

        self.run_name = run_name
        self.save_dir = os.path.join(self.work_dir, run_name)

        if self.with_wandb:
            self.wandb = wandb.init(
                project=self.wandb_project,
                entity=self.wandb_entity,
                dir=self.wandb_dir,
                name=self.run_name,
                save_code=True,
            )

        ## init model, operimizer, loss ...
        self._reset_state()
        self.best_epoch = 0
        self.feature_name = pretrainer.feature_name


    def _init_model_opt(self):
        fixseed(self.seed)
    
        model_args = deepcopy(self.pretrainer.saver.inMemorySave["model_args"])
        model_weights = deepcopy(self.pretrainer.saver.inMemorySave["model_state_dict"])
    
        model_args.update(
            task_dim=self.task_dim,
            task_type=self.task_type,
            task_dense_layer=self.task_dense_layer,
            task_batch_norms=self.task_batch_norms,
            seed=self.seed,
        )
    
        model = Compass(**model_args)
    
        # ===== load pretrained weights =====
        encoder_state = OrderedDict()
        for k, v in model_weights.items():
            if self.load_decoder or "taskdecoder" not in k:
                encoder_state[k] = v
        model.load_state_dict(encoder_state, strict=False)
        model = model.to(self.device)

        # ========= mode setup =========
        if self.mode == "LFT":
            model.train()
            model.inputencoder.apply(freeze_bn_dropout)
            model.latentprojector.apply(freeze_bn_dropout)
            model.taskdecoder.train()
        
            plist = [
                {"params": model.inputencoder.parameters(), "lr": self.lr * 0.1},
                {"params": model.latentprojector.parameters(), "lr": self.lr * 0.1},
                {"params": model.taskdecoder.parameters(), "lr": self.lr},
            ]
        
        elif self.mode == "PFT":
            model.train()
            model.inputencoder.apply(freeze_bn_dropout)
            model.latentprojector.train()
            model.taskdecoder.train()
        
            plist = [
                {"params": model.inputencoder.parameters(), "lr": self.lr * 0.1},
                {"params": model.latentprojector.parameters(), "lr": self.lr},
                {"params": model.taskdecoder.parameters(), "lr": self.lr},
            ]
        
        elif self.mode == "FFT":
            model.train()
            plist = [
                {"params": model.inputencoder.parameters(), "lr": self.lr},
                {"params": model.latentprojector.parameters(), "lr": self.lr},
                {"params": model.taskdecoder.parameters(), "lr": self.lr},
            ]
     
        else:
            raise ValueError("Invalid mode type. Use 'PFT','LFT' or 'FFT'. ")
    
        optimizer = torch.optim.Adam(plist)
        return model, optimizer

    
    def _reset_state(self):
        fixseed(self.seed)
        self.model, self.optimizer = self._init_model_opt()
        self.saver = SaveBestModel(save_dir=self.save_dir, save_name="ft_model.pth")

        ssl_loss = TripletLoss(margin=self.triplet_margin, metric=self.triplet_metric)

        if self.task_loss_type == "ce_loss":
            tsk_loss = CEWithNaNLabelsLoss(weights=self.task_class_weight)
        elif self.task_loss_type == "focal_loss":
            tsk_loss = FocalLoss(weights=self.task_class_weight, gamma=2.0)
        elif self.task_loss_type == "mae_loss":
            tsk_loss = MAEWithNaNLabelsLoss()
        elif self.task_loss_type == "hinge_loss":
            tsk_loss = HingeLoss()
        elif self.task_loss_type == "dice_loss":
            tsk_loss = DiceLoss()
        elif self.task_loss_type == "dsc_loss":
            tsk_loss = DSCLoss()

        self.ssl_loss = ssl_loss
        self.tsk_loss = tsk_loss

    def tune(
        self, dfcx_train, dfy_train, min_mcc=0.8, min_f1=0.8, min_roc=0.8, min_prc=0.8
    ):

        self._reset_state()
        if self.with_wandb:
            self.wandb.watch(self.model)

        dfcx_train = self.scaler.transform(dfcx_train)
        train_itrp = ITRPData(dfcx_train, dfy_train)
        train_loader = data.DataLoader(
            train_itrp,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
            worker_init_fn=worker_init_fn,
            pin_memory=True,
            num_workers=4,
        )  #

        assert train_itrp.y.shape[1] == self.task_dim, "task_dim doesnot matched!"

        # ## init the few-shot taskdecoder based on the representation of the prototypes
        if (self.task_type == "f") & (len(self.task_dense_layer) == 0):
            ## claculate the support set features
            pretrainer = self.pretrainer
            dfcx_train_emb, _ = pretrainer.predict(dfcx_train, batch_size=128)
            proj_feature_names = pretrainer.model.proj_feature_names
            dfcx_train_emb = dfcx_train_emb[proj_feature_names]
            support_features = torch.tensor(dfcx_train_emb.values)
            support_labels = torch.tensor(dfy_train.values)
            self.model.taskdecoder.initialize_parameters(
                support_features, support_labels
            )
            # print(support_features)

        ### training ###
        performance = []
        best_criteria = 1e10
        self.saver(best_criteria, 0, self.model, self.optimizer, self.scaler)

        for epoch in tqdm(range(self.max_epochs), ascii=True):
            train_total_loss, train_ssl_loss, train_tsk_loss = FT_Trainer(
                train_loader,
                self.model,
                self.optimizer,
                self.ssl_loss,
                self.tsk_loss,
                self.device,
                alpha=self.task_loss_weight,
                correction=self.batch_correction,
                entropy_weight=self.entropy_weight,
            )
            train_f1, train_mcc, train_prc, train_roc, train_acc = Evaluator(
                train_loader, self.model, self.device
            )
            performance.append(
                [epoch, train_f1, train_mcc, train_prc, train_roc, train_acc]
            )

            if self.with_wandb:
                self.wandb.log(
                    {
                        "train_f1": train_f1,
                        "train_prc": train_prc,
                        "train_roc": train_roc,
                        "train_acc": train_acc,
                        "train_mcc": train_mcc,
                    }
                )

            if self.verbose:
                print(
                    "Epoch: {}/{} - Train roc: {:.4f}, prc: {:.4f}, f1: {:.4f}, mcc: {:.4f} ".format(
                        epoch + 1,
                        self.max_epochs,
                        train_roc,
                        train_prc,
                        train_f1,
                        train_mcc,
                    )
                )

            # print(train_f1, train_mcc, train_prc, train_roc, train_acc)
            saving_criteria = -train_mcc

            self.saver(saving_criteria, epoch, self.model, self.optimizer, self.scaler)

            if (
                (train_mcc >= min_mcc)
                & (train_f1 >= min_f1)
                & (train_prc >= min_prc)
                & (train_roc >= min_roc)
            ):
                print(
                    "Stopping early at epoch {:2d}. Meet minimal requirements by: f1={:.2f},mcc={:.2f},prc={:.2f}, roc={:.2f}".format(
                        epoch + 1, train_f1, train_mcc, train_prc, train_roc
                    )
                )
                break

        if self.save_best_model:
            self.saver.save()
        else:
            os.system('rm -r "%s"' % self.save_dir)
        self.performance = performance
        self.best_epoch = self.saver.inMemorySave["epoch"]

        ### reload best model weights to model
        self.model.load_state_dict(
            self.saver.inMemorySave["model_state_dict"], strict=True
        )
        self.optimizer.load_state_dict(self.saver.inMemorySave["optimizer_state_dict"])

        return self

    def tune_with_test(
        self,
        dfcx_train,
        dfy_train,
        dfcx_test=None,
        dfy_test=None,
        min_roc=0.8,
        min_prc=0.7,
        min_mcc=0.0,
    ):

        self._reset_state()
        if self.with_wandb:
            self.wandb.watch(self.model)

        ### scaler ####
        dfcx_train = self.scaler.transform(dfcx_train)
        train_itrp = ITRPData(dfcx_train, dfy_train)

        # self.train_itrp = train_itrp
        self.y_scaler = train_itrp.y_scaler
        self.feature_name = train_itrp.feature_name
        train_loader = data.DataLoader(
            train_itrp,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
            worker_init_fn=worker_init_fn,
            pin_memory=True,
            num_workers=4,
        )  #

        assert self.task_dim == train_itrp.y.shape[1], "task_dim doesnot matched!"

        if dfcx_test is not None:
            dfcx_test = self.scaler.transform(dfcx_test)
            test_itrp = ITRPData(dfcx_test, dfy_test)
            test_loader = data.DataLoader(
                test_itrp,
                batch_size=self.batch_size,
                shuffle=False,
                worker_init_fn=worker_init_fn,
                pin_memory=True,
                num_workers=4,
            )
        else:
            test_loader = None

        # ## init the few-shot taskdecoder based on the representation of the prototypes
        if (self.task_type == "f") & (len(self.task_dense_layer) == 0):
            ## claculate the support set features
            pretrainer = self.pretrainer
            dfcx_train_emb, _ = pretrainer.predict(dfcx_train, batch_size=128)
            proj_feature_names = pretrainer.model.proj_feature_names
            dfcx_train_emb = dfcx_train_emb[proj_feature_names]
            support_features = torch.tensor(dfcx_train_emb.values)
            support_labels = torch.tensor(dfy_train.values)
            self.model.taskdecoder.initialize_parameters(
                support_features, support_labels
            )

        ### training ###
        performance = []
        best_val_loss = float("inf")
        patience_counter = 0
        for epoch in tqdm(range(self.max_epochs), ascii=True):
            train_total_loss, train_ssl_loss, train_tsk_loss = FT_Trainer(
                train_loader,
                self.model,
                self.optimizer,
                self.ssl_loss,
                self.tsk_loss,
                self.device,
                alpha=self.task_loss_weight,
                correction=self.batch_correction,
                entropy_weight=self.entropy_weight,
            )

            train_f1, train_mcc, train_prc, train_roc, train_acc = Evaluator(
                train_loader, self.model, self.device
            )
            train_avg_metric = np.mean(
                [train_f1, train_mcc, train_prc, train_roc, train_acc]
            )
            saving_criteria = -(train_prc + train_mcc)
            # print(train_f1, train_mcc, train_prc, train_roc, train_acc)

            if test_loader is not None:
                test_f1, test_mcc, test_prc, test_roc, test_acc = Evaluator(
                    test_loader, self.model, self.device
                )

                test_total_loss, test_ssl_loss, test_tsk_loss = FT_Tester(
                    test_loader,
                    self.model,
                    self.ssl_loss,
                    self.tsk_loss,
                    self.device,
                    alpha=self.task_loss_weight,
                    correction=0.0,
                )

                test_avg_metric = np.mean(
                    [test_f1, test_mcc, test_prc, test_roc, test_acc]
                )
                saving_criteria = -(test_prc + test_mcc)

            else:
                test_total_loss, test_ssl_loss, test_tsk_loss = np.nan, np.nan, np.nan

            ## saving_criteria: lower is better
            ## saving model & Early Stopping based on saving_criteria
            self.saver(saving_criteria, epoch, self.model, self.optimizer, self.scaler)
            if saving_criteria < best_val_loss:
                best_val_loss = saving_criteria
                patience_counter = 0
            else:
                patience_counter += 1

            performance.append(
                [
                    epoch,
                    train_mcc,
                    train_prc,
                    train_tsk_loss,
                    test_mcc,
                    test_prc,
                    test_tsk_loss,
                ]
            )

            if self.with_wandb:
                self.wandb.log(
                    {

                        "FT_train_%s_loss" % self.task_name: train_tsk_loss,
                        "FT_test_%s_loss" % self.task_name: test_tsk_loss,
                    }
                )

            if self.verbose:
                print(
                    "Epoch: {}/{} - Train tsk Loss: {:.4f} - Test tsk Loss: {:.4f}".format(
                        epoch + 1, self.max_epochs, train_tsk_loss, test_tsk_loss
                    )
                )

            if (
                (patience_counter >= self.patience)
                & (train_roc > min_roc)
                & (train_prc > min_prc)
                & (train_mcc > min_mcc)
            ):
                print(
                    f"Stopping early at epoch {epoch+1}. No improvement in validation loss for {self.patience} consecutive epochs."
                )
                break

        self.performance = performance
        self.best_epoch = self.saver.inMemorySave["epoch"]

        ### reload best model weights to model
        self.model.load_state_dict(
            self.saver.inMemorySave["model_state_dict"], strict=True
        )
        self.optimizer.load_state_dict(self.saver.inMemorySave["optimizer_state_dict"])

        if self.save_best_model:
            self.saver.save()
            ## plot loss locally
            df = pd.DataFrame(
                self.performance,
                columns=[
                    "epochs",
                    "train_mcc",
                    "train_prc",
                    "%s_loss" % self.task_name,
                    "test_mcc",
                    "test_prc",
                    "test_%s_loss" % self.task_name,
                ],
            ).set_index("epochs")
            # v = (df - df.min(axis=0)) / (df.max(axis=0) - df.min(axis=0))
            fig, ax = plt.subplots(figsize=(7, 5))
            df.plot(ax=ax)
            fig.savefig(
                os.path.join(self.save_dir, "%s_train_loss.png" % self.task_name),
                bbox_inches="tight",
            )
            plt.close()
            df.to_pickle(
                os.path.join(self.save_dir, "%s_train_loss.pkl" % self.task_name)
            )

        else:
            os.system('rm -r "%s"' % self.save_dir)

        return self

    def predict(self, df_tpm, batch_size=512, num_workers=4):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfe, dfp = Predictor(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        return dfe, dfp

    def extract(self, df_tpm, batch_size=512, num_workers=4, with_gene_level=False):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfg, dfgs, dfct = Extractor(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
            with_gene_level=True,
        )
        if with_gene_level:
            return dfg, dfgs, dfct
        else:
            return dfgs, dfct

    def project(self, df_tpm, batch_size=512, num_workers=4):
        model = Compass(**self.saver.inMemorySave["model_args"])
        model.load_state_dict(self.saver.inMemorySave["model_state_dict"], strict=False)
        model = model.to(self.device)
        dfg, dfc = Projector(
            df_tpm,
            model,
            self.scaler,
            device=self.device,
            batch_size=batch_size,
            num_workers=num_workers,
        )
        return dfg, dfc

    def plot_embed(self, df_tpm, df_label, label_types, **kwargs):

        ### make prediction & plot on TCGA
        dfe, dfp = self.predict(df_tpm, batch_size=self.batch_size, num_workers=4)
        dfd = dfe.join(df_label)
        label_cols = df_label.columns

        figs = plot_embed_with_label(
            dfd, label_col=label_cols, label_type=label_types, **kwargs
        )
        for fig, name in zip(figs, label_cols):
            fig.savefig(
                os.path.join(self.save_dir, "FT_%s.png" % name),
                bbox_inches="tight",
            )

    def save(self, mfile):
        if self.with_wandb:
            self.wandb._settings = ""
        torch.save(self, mfile)
        print("Saving the model to %s" % mfile)

    def load(self, mfile, **kwargs):
        self = loadcompass(mfile, **kwargs)
        if self.with_wandb:
            self.wandb._settings = ""
        return self

    def close(self):
        if self.with_wandb:
            self.wandb.finish()
            self.wandb._settings = ""
        self.save(os.path.join(self.save_dir, "finetuner.pt"))

    def copy(self):
        return deepcopy(self)

    def count_parameters(self):
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)

    def get_tuner_epoch(
        self,
        train_X,
        train_y,
        max_epochs=200,
        work_dir="./Paramtune",
        with_wandb=False,
        verbose=False,
        fold=3,
    ):

        param = deepcopy(self.params)
        param["with_wandb"] = with_wandb
        param["epochs"] = max_epochs
        param["work_dir"] = work_dir
        param["verbose"] = verbose

        skf = StratifiedKFold(n_splits=fold, shuffle=True, random_state=42)
        splits = skf.split(train_y.values, train_y.idxmax(axis=1))

        best_epochs = []
        for i, (train_idx, valid_idx) in enumerate(splits):
            X_train = train_X.iloc[train_idx]
            y_train = train_y.iloc[train_idx]
            X_val = train_X.iloc[valid_idx]
            y_val = train_y.iloc[valid_idx]
            print(
                len(X_train),
                len(X_val),
                y_val.idxmax(axis=1).value_counts().to_dict(),
                y_train.idxmax(axis=1).value_counts().to_dict(),
            )

            tuner = FineTuner(
                self.pretrainer,
                **param,
                task_name="fold_%s" % i,
                task_type="c",
            )

            tuner.tune(
                dfcx_train=X_train, dfy_train=y_train, dfcx_test=X_val, dfy_test=y_val
            )

            best_epoch = tuner.saver.inMemorySave["epoch"]
            best_epochs.append(best_epoch)

        print(np.array(best_epochs))
        best_epochs = int(np.array(best_epochs).mean())

        return best_epochs



import torch
import torch.nn as nn
import torch.nn.functional as F


class GeneSetScorePooling(nn.Module):
    def __init__(self, pooling_type="mean"):
        """
        Initializes the GeneSetScorePooling module for gene set score calculation.
        :param pooling_type: A string indicating the type of pooling ('mean' or 'max').
        """
        super(GeneSetScorePooling, self).__init__()
        self.pooling_type = pooling_type

    def forward(self, x):
        """
        Forward pass of the module.
        :param x: A tensor of shape (batch_size, num_gene_sets, feature_dim).
        :return: A tensor of shape (batch_size, num_gene_sets).
        """
        if self.pooling_type == "mean":
            return torch.mean(x, dim=2)
        elif self.pooling_type == "max":
            return torch.max(x, dim=2)[0]
        else:
            raise ValueError("Invalid pooling type. Use 'mean' or 'max'.")


class DRNet(nn.Module):
    def __init__(self, feature_dim):
        super(DRNet, self).__init__()
        self.fc1 = nn.Linear(feature_dim, 16)
        self.bn1 = nn.BatchNorm1d(16)
        self.fc2 = nn.Linear(16, 1)

    def forward(self, x):
        x = F.relu(self.bn1(self.fc1(x)))  #
        x = self.fc2(x)
        return x


class GeneSetScoreLinear(nn.Module):
    def __init__(self, feature_dim):
        """
        Initializes the GeneSetScoreLinear module for gene set score calculation.
        :param feature_dim: The dimension of features for each gene set.
        """
        super(GeneSetScoreLinear, self).__init__()
        self.fc = nn.Linear(
            feature_dim, 1
        )  # Linear layer to transform features to a single score
        # self.fc = DRNet(feature_dim)

    def forward(self, x):
        """
        Forward pass of the module.
        :param x: A tensor of shape (batch_size, num_gene_sets, feature_dim).
        :return: A tensor of shape (batch_size, num_gene_sets).
        """
        batch_size, num_gene_sets, _ = x.shape
        x = x.view(-1, x.size(-1))  # Flatten the last two dimensions for linear layer
        scores = self.fc(x)
        # scores = F.relu(scores) # score greater than zero
        scores = scores.view(
            batch_size, num_gene_sets
        )  # Reshape to original batch and gene set dimensions
        return scores



class GeneSetScorer(nn.Module):
    def __init__(self, feature_dim, mode="linear", pooling_type="mean"):
        """
        Initializes the GeneSetScore module for gene set score calculation.
        :param mode: A string indicating the type of pooling ('mean' or 'max', or 'linear').

        # Example usage
        ==================
        >>> gene_set_level_features = torch.rand(10, 3, 32)  # Assume gene_set_level_features has shape
        >>> scorer = GeneSetScorer(32, mode='linear')
        >>> scores = scorer(gene_set_level_features)  # shape will be (256, 3)
        >>> scores.shape
        """

        super(GeneSetScorer, self).__init__()
        self.mode = mode
        self.feature_dim = feature_dim

        if mode == "pooling":
            self.genesetscorer = GeneSetScorePooling(pooling_type)
        elif mode == "linear":
            self.genesetscorer = GeneSetScoreLinear(feature_dim)
        else:
            raise ValueError("Invalid mode type. Use 'mean','max' or 'linear'. ")

    def forward(self, x):
        scores = self.genesetscorer(x)
        return scores



import numpy as np
import torch
from sklearn.preprocessing import MinMaxScaler
from torch.utils.data import Dataset
from torch.nn.functional import normalize
from torch.distributions import Beta


class MixupNormalAugmentor:
    """Mixup from normal samples with selected genes"""

    def __init__(self, df_tpm_normal, genes2mixup=[], mix_beta=0.9, n_views=1):
        """Selective mixup initialization."""
        if len(genes2mixup) == 0:
            genes2mixup = df_tpm_normal.columns

        X = torch.tensor(df_tpm_normal.values, dtype=torch.float32).clone().detach()

        self.genes2mixup = genes2mixup
        self.select_idx = df_tpm_normal.columns.get_indexer(genes2mixup)
        self.X_mix = X[:, self.select_idx]

        self.n = len(df_tpm_normal)
        self.beta = mix_beta
        self.p = Beta(torch.FloatTensor([mix_beta]), torch.FloatTensor([mix_beta]))
        self.n_views = n_views

    def _transform(self, x):
        b1 = self.beta
        m1 = self.X_mix[np.random.choice(self.n)]
        x1 = x.clone().detach()
        xs = x1[self.select_idx]
        xs = b1 * xs + (1 - b1) * m1
        x1[self.select_idx] = xs
        return x1.to(x.device)

    def augment(self, x):
        return [self._transform(x) for _ in range(self.n_views)]

    def __repr__(self):
        return f"{self.__class__.__name__}(beta={self.beta}, n_views={self.n_views})"


class RandomMaskAugmentor:
    """Random Mask Augmentation for Gene Expression Vectors."""

    def __init__(
        self,
        mask_p_prob=0.1,
        mask_a_prob=None,
        mask_n_prob=None,
        no_augment_prob=0.1,
        n_views=1,
    ):
        if mask_a_prob is None:
            mask_a_prob = mask_p_prob
        if mask_n_prob is None:
            mask_n_prob = mask_p_prob

        self.mask_p_prob = mask_p_prob
        self.mask_a_prob = mask_a_prob
        self.mask_n_prob = mask_n_prob
        self.no_augment_prob = no_augment_prob
        self.n_views = n_views

    def _transform(self, x, probability):
        if torch.rand(1).item() < self.no_augment_prob:
            return x.clone()

        random_values = torch.rand(len(x[1:]), device=x.device)
        mask = random_values < probability

        x_new = x.clone()
        x_new[1:][mask] = 0
        return x_new

    def _augment(self, x, mask_prob):
        assert len(x.shape) == 1, "augment for each sample on a vector!"
        return [self._transform(x, mask_prob) for _ in range(self.n_views)]

    def augment_p(self, p):
        return self._augment(p, self.mask_p_prob)

    def augment_a(self, a):
        return self._augment(a, self.mask_a_prob)

    def augment_n(self, n):
        return self._augment(n, self.mask_n_prob)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(mask_probability=(a:{self.mask_a_prob}, "
            f"p:{self.mask_p_prob}, n:{self.mask_n_prob}), "
            f"no={self.no_augment_prob}, n_views={self.n_views})"
        )


class FeatureJitterAugmentor:
    """Feature Jittering for Gene Expression Vectors."""

    def __init__(
        self,
        jitter_p_std=0.1,
        jitter_a_std=None,
        jitter_n_std=None,
        no_augment_prob=0.1,
        n_views=1,
    ):
        if jitter_a_std is None:
            jitter_a_std = jitter_p_std
        if jitter_n_std is None:
            jitter_n_std = jitter_p_std

        self.n_views = n_views
        self.jitter_p_std = jitter_p_std
        self.jitter_a_std = jitter_a_std
        self.jitter_n_std = jitter_n_std
        self.no_augment_prob = no_augment_prob

    def _transform(self, x, jitter_std):
        if torch.rand(1) < self.no_augment_prob:
            return x.clone()

        jitter = torch.normal(
            mean=0, std=jitter_std, size=x[1:].size(), device=x.device
        )
        x_jittered = x.clone()
        x_jittered[1:] += jitter
        return x_jittered

    def _augment(self, x, jitter_std):
        assert len(x.shape) == 1, "augment for each sample on a vector!"
        return [self._transform(x, jitter_std) for _ in range(self.n_views)]

    def augment_p(self, p):
        return self._augment(p, self.jitter_p_std)

    def augment_a(self, a):
        return self._augment(a, self.jitter_a_std)

    def augment_n(self, n):
        return self._augment(n, self.jitter_n_std)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(jitter_std=(a:{self.jitter_a_std}, "
            f"p:{self.jitter_p_std}, n:{self.jitter_n_std}, "
            f"no:{self.no_augment_prob}), n_views={self.n_views})"
        )


class MaskJitterAugmentor:
    """Mixed augmentation (Masking and Jittering)."""

    def __init__(
        self,
        mask_p_prob=0.01,
        mask_a_prob=None,
        mask_n_prob=0.0,
        jitter_p_std=0.01,
        jitter_a_std=None,
        jitter_n_std=0.0,
        no_augment_prob=0.1,
        n_views=1,
    ):
        self.mask_p_prob = mask_p_prob
        self.mask_a_prob = mask_a_prob
        self.mask_n_prob = mask_n_prob
        self.jitter_p_std = jitter_p_std
        self.jitter_a_std = jitter_a_std
        self.jitter_n_std = jitter_n_std
        self.no_augment_prob = no_augment_prob
        self.n_views = n_views

        self.augmentor1 = RandomMaskAugmentor(
            mask_p_prob, mask_a_prob, mask_n_prob, no_augment_prob, n_views
        )
        self.augmentor2 = FeatureJitterAugmentor(
            jitter_p_std, jitter_a_std, jitter_n_std, no_augment_prob, n_views
        )

    def augment_p(self, p):
        augmentor = np.random.choice([self.augmentor1, self.augmentor2])
        return augmentor.augment_p(p)

    def augment_a(self, a):
        augmentor = np.random.choice([self.augmentor1, self.augmentor2])
        return augmentor.augment_a(a)

    def augment_n(self, n):
        augmentor = np.random.choice([self.augmentor1, self.augmentor2])
        return augmentor.augment_n(n)



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 16:44:09 2023

@author: Wanxiang Shen
"""
import torch
import torch.utils.data as Torchdata
import pandas as pd
import numpy as np
from tqdm import tqdm

tqdm.pandas(ascii=True)

from ..dataloader import GeneData
from .loss import entropy_regularization, independence_loss
from .loss import cv_loss, msd_loss, cv_loss_penalty


import gc


def worker_init_fn(worker_id):
    seed = torch.initial_seed() % 2**32
    np.random.seed(seed)


def PT_Trainer(
    train_loader,
    model,
    optimizer,
    ssl_loss,
    tsk_loss,
    device,
    alpha=0.0,
    correction=0.0,
    entropy_weight=0.0,
):

    model.train()
    total_loss = []
    total_ssl_loss = []
    total_tsk_loss = []

    # torch.autograd.set_detect_anomaly(True)
    # for data in tqdm(train_loader, ascii=True):
    for data in train_loader:

        triplet, label = data

        anchor_y_true, positive_y_true, negative_y_true = label
        anchor, positive, negative = triplet

        anchor, positive, negative = (
            anchor.to(device),
            positive.to(device),
            negative.to(device),
        )
        anchor_y_true = anchor_y_true.to(device)
        positive_y_true = positive_y_true.to(device)
        negative_y_true = negative_y_true.to(device)

        optimizer.zero_grad()

        (anchor_emb, anchor_refg), anchor_y_pred = model(anchor)
        (positive_emb, positive_refg), positive_y_pred = model(positive)
        (negative_emb, negative_refg), negative_y_pred = model(negative)

        lss = ssl_loss(anchor_emb, positive_emb, negative_emb)

        ## remove batch effects by minimal the differences between house-keeping genes
        if correction != 0:
            refe = torch.cat(
                [anchor_refg[1], positive_refg[1], negative_refg[1]], axis=0
            )
            refy = torch.cat([anchor_y_true, positive_y_true, negative_y_true], axis=0)
            # ref = cv_loss_penalty(refe)
            ref = msd_loss(refe)
            # idp = independence_loss(refe, refy)
            # print("Lss: {:.3f} - Ref: {:.3f} - idp: {:.6f}".format(lss.item(), ref.item(), idp.item()))
            lss = (1 - correction) * lss + correction * ref
        y_pred = anchor_y_pred  # torch.cat([anchor_y_pred, positive_y_pred, negative_y_pred])
        y_true = anchor_y_true  # torch.cat([anchor_y_true, positive_y_true, negative_y_true])
        tsk = tsk_loss(y_pred, y_true)

        if entropy_weight != 0:
            entropy_reg = entropy_regularization(y_pred)
            tsk = tsk * (1 - entropy_weight) + entropy_reg * entropy_weight

        loss = (1.0 - alpha) * lss + tsk * alpha

        # print(cv_loss, lss, loss)
        loss.backward()
        optimizer.step()

        total_loss.append(loss.item())
        total_ssl_loss.append(lss.item())
        total_tsk_loss.append(tsk.item())

        torch.cuda.empty_cache()
        gc.collect()

    train_total_loss = np.mean(total_loss)
    train_ssl_loss = np.mean(total_ssl_loss)
    train_tsk_loss = np.mean(total_tsk_loss)

    return train_total_loss, train_ssl_loss, train_tsk_loss


@torch.no_grad()
def PT_Tester(test_loader, model, ssl_loss, tsk_loss, device, alpha=0.0, correction=0):
    model.eval()
    total_loss = []
    total_ssl_loss = []
    total_tsk_loss = []

    for data in test_loader:
        triplet, label = data
        anchor, positive, negative = triplet
        anchor_y_true, positive_y_true, negative_y_true = label

        anchor = anchor.to(device)
        positive = positive.to(device)
        negative = negative.to(device)
        anchor_y_true = anchor_y_true.to(device)
        positive_y_true = positive_y_true.to(device)
        negative_y_true = negative_y_true.to(device)

        (anchor_emb, anchor_refg), anchor_y_pred = model(anchor)

        (positive_emb, positive_refg), positive_y_pred = model(positive)
        (negative_emb, negative_refg), negative_y_pred = model(negative)

        lss = ssl_loss(anchor_emb, positive_emb, negative_emb)

        if correction != 0:
            refe = torch.cat(
                [anchor_refg[1], positive_refg[1], negative_refg[1]], axis=0
            )
            refy = torch.cat([anchor_y_true, positive_y_true, negative_y_true], axis=0)
            # ref = cv_loss(refe) + msd_loss(refe)
            ref = msd_loss(refe)  #
            lss = (1 - correction) * lss + correction * ref
        y_pred = anchor_y_pred
        y_true = anchor_y_true
        tsk = tsk_loss(y_pred, y_true)

        loss = (1.0 - alpha) * lss + tsk * alpha

        total_loss.append(loss.item())
        total_ssl_loss.append(lss.item())
        total_tsk_loss.append(tsk.item())

    test_total_loss = np.mean(total_loss)
    test_ssl_loss = np.mean(total_ssl_loss)
    test_tsk_loss = np.mean(total_tsk_loss)

    return test_total_loss, test_ssl_loss, test_tsk_loss



import pandas as pd
from itertools import chain
import numpy as np
import os, json
import torch

cwd = os.path.dirname(__file__)


with open(os.path.join(cwd, "cancer_code.json"), "r") as file:
    CANCER_CODE = json.load(file)

with open(os.path.join(cwd, "gene_tokens_long.json"), "r") as file:
    TOKENS_LONG = json.load(file)

with open(os.path.join(cwd, "gene_tokens_short.json"), "r") as file:
    TOKENS_SHORT = json.load(file)

CONCEPT = pd.read_csv(
    os.path.join(cwd, "conception_processed.tsv"), index_col=0, sep="\t"
)  # _processed


dfd = pd.read_excel(os.path.join(cwd, "concept_colors.xlsx"))
CONCEPT_palette = dfd.set_index("Concept_name").Color.to_dict()


RESPONSE_palette = {"R": "#008bfb", "NR": "#ff0051"}  # ,





import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn import MultiheadAttention, Dropout, LayerNorm
from typing import Optional, Callable, Union
from abc import ABC, abstractmethod


# 自定义线性层，包括谱归一化
class Linear(nn.Linear):
    # 初始化函数
    def __init__(self, in_features, out_features, bias=True, device=None, dtype=None):
        super().__init__(
            in_features, out_features, bias=bias, device=device, dtype=dtype
        )
        # 初始化谱归一化
        self.register_buffer(
            "u", nn.functional.normalize(torch.randn(in_features), dim=0)
        )
        sigma = self.get_sigma()
        self.register_buffer("spectral_norm", sigma)
        self.sigma = nn.Parameter(torch.ones(1))

    # 获取谱归一化的 sigma 值
    def get_sigma(self):
        with torch.no_grad():
            u = self.u
            v = self.weight.mv(u)
            v = nn.functional.normalize(v, dim=0)
            u = self.weight.T.mv(v)
            u = nn.functional.normalize(u, dim=0)
            self.u.data.copy_(u)
        return torch.einsum("c,cd,d->", v, self.weight, u)

    # 重写 forward 方法，应用谱归一化
    def forward(self, x):
        weight = (self.sigma / self.spectral_norm) * self.weight
        return nn.functional.linear(x, weight, self.bias)


# 获取激活函数
def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu
    raise RuntimeError(f"activation should be relu/gelu, not {activation}")


# 抽象 Transformer 层基类
class AbstractTrasnformerLayer(ABC):
    @abstractmethod
    def __init__(
        self, embed_dim, num_heads, dropout, norm, norm_first: bool, causal: bool
    ):
        pass

    @abstractmethod
    def forward(self, x, attn_mask, output_attentions):
        pass


# 标准 Transformer 编码器层
class TransformerEncoderLayer(nn.Module):
    # 初始化函数
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        device=None,
        dtype=None,
    ) -> None:
        super().__init__()
        factory_kwargs = {"device": device, "dtype": dtype}
        self.self_attn = MultiheadAttention(
            d_model, nhead, dropout=dropout, batch_first=batch_first, **factory_kwargs
        )
        self.linear1 = Linear(d_model, dim_feedforward, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = Linear(dim_feedforward, d_model, **factory_kwargs)
        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.activation = (
            _get_activation_fn(activation)
            if isinstance(activation, str)
            else activation
        )

    # 前向传播函数
    def forward(
        self,
        src: Tensor,
        src_mask: Optional[Tensor] = None,
        src_key_padding_mask: Optional[Tensor] = None,
    ) -> Tensor:
        x = src
        if self.norm_first:
            x = x + self._sa_block(self.norm1(x), src_mask, src_key_padding_mask)
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(x + self._sa_block(x, src_mask, src_key_padding_mask))
            x = self.norm2(x + self._ff_block(x))
        return x

    # 自注意力 block
    def _sa_block(
        self, x: Tensor, attn_mask: Optional[Tensor], key_padding_mask: Optional[Tensor]
    ) -> Tensor:
        x, _ = self.self_attn(
            x, x, x, attn_mask=attn_mask, key_padding_mask=key_padding_mask
        )
        return self.dropout1(x)

    # 前馈网络 block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)


# Vanilla Transformer 层
class VanillaTransformerLayer(nn.Module, AbstractTrasnformerLayer):
    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        norm="layernorm",
        norm_first=True,
        causal=False,
    ):
        super().__init__()
        assert norm == "layernorm", "Vanilla transformer only supports layernorm."
        assert not causal, "Vanilla transformer does not support causal inference."
        self.layer = TransformerEncoderLayer(
            embed_dim,
            num_heads,
            embed_dim * 2,
            dropout,
            activation="gelu",
            norm_first=norm_first,
        )
        self.support_output_attentions = False

    def forward(self, x, attn_mask=None, output_attentions=False):
        assert (
            not output_attentions
        ), "output_attentions not implemented for VanillaTransformer"
        return self.layer(x, attn_mask)
